#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_coords  # noqa: E402
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units",
    [
        (
            {
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {"date": [20240603, 20240604], "time": [0, 1200], "step": [0, 6]},
            ("step", "hours"),
        ),
        (
            {
                "time_dim_mode": "raw",
                "dim_name_from_role_name": True,
            },
            {
                "date": [np.datetime64("2024-06-03", "ns"), np.datetime64("2024-06-04", "ns")],
                "time": [np.timedelta64(0, "s"), np.timedelta64(43200, "s")],
                "step": [np.timedelta64(0, "h"), np.timedelta64(6, "h")],
            },
            None,
        ),
        (
            {
                "time_dim_mode": "forecast",
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                    np.datetime64("2024-06-04T00", "ns"),
                    np.datetime64("2024-06-04T12", "ns"),
                ],
                "step": [0, 6],
            },
            ("step", "hours"),
        ),
        (
            {
                "time_dim_mode": "forecast",
                "dim_name_from_role_name": True,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                    np.datetime64("2024-06-04T00", "ns"),
                    np.datetime64("2024-06-04T12", "ns"),
                ],
                "step": [np.timedelta64(0, "h"), np.timedelta64(6, "h")],
            },
            None,
        ),
        (
            {
                "time_dim_mode": "valid_time",
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "valid_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T06", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                    np.datetime64("2024-06-03T18", "ns"),
                    np.datetime64("2024-06-04T00", "ns"),
                    np.datetime64("2024-06-04T06", "ns"),
                    np.datetime64("2024-06-04T12", "ns"),
                    np.datetime64("2024-06-04T18", "ns"),
                ],
            },
            None,
        ),
        (
            {
                "time_dim_mode": "valid_time",
                "decode_times": True,
                "decode_timedelta": True,
                "dim_name_from_role_name": True,
            },
            {
                "valid_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T06", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                    np.datetime64("2024-06-03T18", "ns"),
                    np.datetime64("2024-06-04T00", "ns"),
                    np.datetime64("2024-06-04T06", "ns"),
                    np.datetime64("2024-06-04T12", "ns"),
                    np.datetime64("2024-06-04T18", "ns"),
                ],
            },
            None,
        ),
        (
            {
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {"date": [20240603, 20240604], "time": [0, 1200], "step_timedelta": [0, 6]},
            ("step_timedelta", "hours"),
        ),
        (
            {
                "time_dim_mode": "raw",
                "dim_name_from_role_name": False,
            },
            {
                "date": [np.datetime64("2024-06-03", "ns"), np.datetime64("2024-06-04", "ns")],
                "time": [np.timedelta64(0, "s"), np.timedelta64(43200, "s")],
                "step_timedelta": [np.timedelta64(0, "h"), np.timedelta64(6, "h")],
            },
            None,
        ),
    ],
)
def test_xr_time_basic(kwargs, dims, step_units):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="t")

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units",
    [
        (
            {
                "dim_roles": {"date": "indexingDate", "time": "indexingTime", "step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {
                "indexing_time": [
                    np.datetime64("2014-09-01", "ns"),
                    np.datetime64("2014-10-01", "ns"),
                ],
                "forecastMonth": [1, 2, 3],
            },
            ("forecastMonth", "months"),
        ),
        (
            {
                "dim_roles": {"forecast_reference_time": "indexing_time", "step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {
                "indexing_time": [
                    np.datetime64("2014-09-01", "ns"),
                    np.datetime64("2014-10-01", "ns"),
                ],
                "forecastMonth": [1, 2, 3],
            },
            ("forecastMonth", "months"),
        ),
        (
            {
                "dim_roles": {"date": "indexingDate", "time": "indexingTime", "step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2014-09-01", "ns"),
                    np.datetime64("2014-10-01", "ns"),
                ],
                "step": [1, 2, 3],
            },
            ("step", "months"),
        ),
        (
            {
                "dim_roles": {"forecast_reference_time": "indexing_time", "step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2014-09-01", "ns"),
                    np.datetime64("2014-10-01", "ns"),
                ],
                "step": [1, 2, 3],
            },
            ("step", "months"),
        ),
    ],
)
def test_xr_time_seasonal_monthly_indexing_date(kwargs, dims, step_units):
    ds_ek = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/xr_engine/date/jma_seasonal_fc_ref_time_per_member.grib"),
    )

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="2t")

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units",
    [
        (
            {
                "time_dim_mode": "forecast",
                "dim_roles": {"step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {
                "number": [0, 1, 2],
                "forecast_reference_time": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "forecastMonth": [1, 2, 3, 4, 5, 6],
            },
            ("forecastMonth", "months"),
        ),
        (
            {
                "time_dim_mode": "forecast",
                "dim_roles": {"step": "fcmonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {
                "number": [0, 1, 2],
                "forecast_reference_time": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "fcmonth": [1, 2, 3, 4, 5, 6],
            },
            ("fcmonth", "months"),
        ),
        (
            {
                "time_dim_mode": "raw",
                "dim_roles": {"step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "ensure_dims": ["number", "date", "time", "forecastMonth"],
                "dim_name_from_role_name": False,
            },
            {
                "number": [0, 1, 2],
                "date": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "time": [np.timedelta64(0, "s")],
                "forecastMonth": [1, 2, 3, 4, 5, 6],
            },
            ("forecastMonth", "months"),
        ),
        (
            {
                "time_dim_mode": "forecast",
                "dim_roles": {"step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "number": [0, 1, 2],
                "forecast_reference_time": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "step": [1, 2, 3, 4, 5, 6],
            },
            ("step", "months"),
        ),
        (
            {
                "time_dim_mode": "forecast",
                "dim_roles": {"step": "fcmonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "number": [0, 1, 2],
                "forecast_reference_time": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "step": [1, 2, 3, 4, 5, 6],
            },
            ("step", "months"),
        ),
        (
            {
                "time_dim_mode": "raw",
                "dim_roles": {"step": "forecastMonth"},
                "decode_times": False,
                "decode_timedelta": False,
                "ensure_dims": ["number", "date", "time", "step"],
                "dim_name_from_role_name": True,
            },
            {
                "number": [0, 1, 2],
                "date": [
                    np.datetime64("1993-10-01", "ns"),
                    np.datetime64("1994-10-01", "ns"),
                    np.datetime64("1995-10-01", "ns"),
                    np.datetime64("1996-10-01", "ns"),
                ],
                "time": [np.timedelta64(0, "s")],
                "step": [1, 2, 3, 4, 5, 6],
            },
            ("step", "months"),
        ),
    ],
)
def test_xr_time_seasonal_monthly_simple(kwargs, dims, step_units):
    ds_ek = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/xr_engine/date/seasonal_monthly.grib"),
    )

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="2t")

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units,coords",
    [
        (
            {
                "time_dim_mode": "forecast",
                "add_valid_time_coord": True,
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": True,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                ],
                "step": [0, 6],
            },
            ("step", "hours"),
            {
                "valid_time": [
                    [np.datetime64("2024-06-03T00", "ns"), np.datetime64("2024-06-03T06", "ns")],
                    [np.datetime64("2024-06-03T12", "ns"), np.datetime64("2024-06-03T18", "ns")],
                ]
            },
        ),
        (
            {
                "fixed_dims": ["level", "forecast_reference_time", "step"],
                "add_valid_time_coord": True,
                "decode_times": False,
                "decode_timedelta": False,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                ],
                "step": [0, 6],
            },
            ("step", "hours"),
            {
                "valid_time": [
                    [np.datetime64("2024-06-03T00", "ns"), np.datetime64("2024-06-03T06", "ns")],
                    [np.datetime64("2024-06-03T12", "ns"), np.datetime64("2024-06-03T18", "ns")],
                ]
            },
        ),
        (
            {
                "fixed_dims": ["level", "step", "forecast_reference_time"],
                "add_valid_time_coord": True,
                "decode_times": False,
                "decode_timedelta": False,
            },
            {
                "step": [0, 6],
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                ],
            },
            ("step", "hours"),
            {
                "valid_time": [
                    [np.datetime64("2024-06-03T00", "ns"), np.datetime64("2024-06-03T12", "ns")],
                    [np.datetime64("2024-06-03T06", "ns"), np.datetime64("2024-06-03T18", "ns")],
                ]
            },
        ),
        (
            {
                "time_dim_mode": "forecast",
                "add_valid_time_coord": True,
                "decode_times": False,
                "decode_timedelta": False,
                "dim_name_from_role_name": False,
            },
            {
                "forecast_reference_time": [
                    np.datetime64("2024-06-03T00", "ns"),
                    np.datetime64("2024-06-03T12", "ns"),
                ],
                "step_timedelta": [0, 6],
            },
            ("step_timedelta", "hours"),
            {
                "valid_time": [
                    [np.datetime64("2024-06-03T00", "ns"), np.datetime64("2024-06-03T06", "ns")],
                    [np.datetime64("2024-06-03T12", "ns"), np.datetime64("2024-06-03T18", "ns")],
                ]
            },
        ),
    ],
)
def test_xr_valid_time_coord(kwargs, dims, step_units, coords):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl_small.grib")).sel(
        date=20240603, time=[0, 1200]
    )

    ds = ds_ek.to_xarray(**kwargs)

    compare_dims(ds, dims, order_ref_var="t")

    vt = ds.coords["valid_time"]
    assert vt.dims == tuple(dims.keys())

    compare_coords(ds, coords)

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units",
    [
        (
            {
                "time_dim_mode": "raw",
                "dim_name_from_role_name": True,
                "ensure_dims": ["date", "time", "step"],
            },
            {
                "date": [np.datetime64("2011-12-15", "ns")],
                "time": [np.timedelta64(12, "h")],
                "step": [
                    np.timedelta64(12, "h"),
                    np.timedelta64(18, "h"),
                    np.timedelta64(24, "h"),
                    np.timedelta64(30, "h"),
                    np.timedelta64(36, "h"),
                ],
            },
            None,
        ),
    ],
)
def test_xr_time_step_range_1(kwargs, dims, step_units):
    ds_ek = from_source(
        "url", earthkit_remote_test_data_file("test-data/xr_engine/date/wgust_step_range.grib1")
    )

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="10fg6")

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dims,step_units",
    [
        (
            {
                "time_dim_mode": "raw",
                "dim_name_from_role_name": True,
                "ensure_dims": ["date", "time", "step"],
            },
            {
                "date": [np.datetime64("2025-05-27", "ns")],
                "time": [np.timedelta64(0, "ns")],
                "step": [np.timedelta64(72, "h"), np.timedelta64(73, "h")],
            },
            None,
        ),
    ],
)
def test_xr_time_step_range_2(kwargs, dims, step_units):
    ds_ek = from_source(
        "url", earthkit_remote_test_data_file("test-data/xr_engine/date/lsp_step_range.grib2")
    )

    ds = ds_ek.to_xarray(**kwargs)
    compare_dims(ds, dims, order_ref_var="lsp")

    if step_units is not None:
        assert (
            ds[step_units[0]].attrs["units"] == step_units[1]
        ), f"step units mismatch {ds[step_units[0]].attrs['units']} != {step_units[1]}"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data import to_target
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_remote_test_data_file


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "raw"},
        {"profile": "mars", "time_dim_mode": "forecast"},
        {"profile": "mars", "time_dim_mode": "raw", "decode_times": False, "decode_timedelta": False},
        {"profile": "mars", "time_dim_mode": "forecast", "decode_times": False, "decode_timedelta": False},
    ],
)
def test_xr_write_1(kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy()
    ref_r_vals = ds_ek.sel(param="r", step=6, level=500).to_numpy()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # data-array
    r = ds["t"].earthkit.to_fieldlist()
    assert len(r) == 16
    assert r.index("shortName") == ["t"]
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())

    ref_base = [
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T12:00:00",
        "2024-06-03T12:00:00",
        "2024-06-03T12:00:00",
        "2024-06-03T12:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T12:00:00",
        "2024-06-04T12:00:00",
        "2024-06-04T12:00:00",
        "2024-06-04T12:00:00",
    ]

    ref_valid = [
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T06:00:00",
        "2024-06-03T06:00:00",
        "2024-06-03T12:00:00",
        "2024-06-03T12:00:00",
        "2024-06-03T18:00:00",
        "2024-06-03T18:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T00:00:00",
        "2024-06-04T06:00:00",
        "2024-06-04T06:00:00",
        "2024-06-04T12:00:00",
        "2024-06-04T12:00:00",
        "2024-06-04T18:00:00",
        "2024-06-04T18:00:00",
    ]

    assert r.metadata("base_datetime") == ref_base
    assert r.metadata("valid_datetime") == ref_valid

    # dataset
    r = ds.earthkit.to_fieldlist()
    assert len(r) == 16 * 2
    assert set(r.index("shortName")) == set(["t", "r"])
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())
    assert np.allclose(ref_r_vals + 1.0, r.sel(param="r", step=6, level=500).to_numpy())

    assert sorted(r.metadata("base_datetime")) == sorted(ds_ek.metadata("base_datetime"))
    assert sorted(r.metadata("valid_datetime")) == sorted(ds_ek.metadata("valid_datetime"))


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "valid_time"},
        {"profile": "mars", "time_dim_mode": "valid_time", "decode_times": False, "decode_timedelta": False},
    ],
)
def test_xr_write_2(kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(date=20240603, time=0, param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy()
    ref_r_vals = ds_ek.sel(param="r", step=6, level=500).to_numpy()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    # NOTE: the basetime and step are lost when using valid_time dim
    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # TODO: currently base_time + step is lost when valid_time dim is used
    # Once we have a solution for this, we need to update the test

    # data-array
    r = ds["t"].earthkit.to_fieldlist()
    assert len(r) == 4
    assert r.index("shortName") == ["t"]
    assert r.index("step") == [0]
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", time=600, level=500).to_numpy())

    ref_base = [
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T06:00:00",
        "2024-06-03T06:00:00",
    ]

    assert r.metadata("base_datetime") == ref_base
    assert r.metadata("valid_datetime") == ref_base

    # # dataset
    r = ds.earthkit.to_fieldlist()
    assert len(r) == 4 * 2
    assert set(r.index("shortName")) == set(["t", "r"])
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", time=600, level=500).to_numpy())
    assert np.allclose(ref_r_vals + 1.0, r.sel(param="r", time=600, level=500).to_numpy())

    assert sorted(r.metadata("base_datetime")) == sorted(ds_ek.metadata("valid_datetime"))
    assert sorted(r.metadata("valid_datetime")) == sorted(ds_ek.metadata("valid_datetime"))


@pytest.mark.cache
def test_xr_write_level_and_type():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(date=20240603, time=0, param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy()
    ref_r_vals = ds_ek.sel(param="r", step=6, level=500).to_numpy()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(level_dim_mode="level_and_type")
    ds += 1

    # TODO: currently base_time + step is lost when valid_time dim is used
    # Once we have a solution for this, we need to update the test

    # data-array
    r = ds["t"].earthkit.to_fieldlist()
    assert len(r) == 4
    assert r.index("shortName") == ["t"]
    assert r.index("step") == [0, 6]
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())

    ref_base = ["2024-06-03T00:00:00"] * 4
    ref_valid = [
        "2024-06-03T00:00:00",
        "2024-06-03T00:00:00",
        "2024-06-03T06:00:00",
        "2024-06-03T06:00:00",
    ]

    assert r.metadata("base_datetime") == ref_base
    assert r.metadata("valid_datetime") == ref_valid

    # dataset
    r = ds.earthkit.to_fieldlist()
    assert len(r) == 4 * 2
    assert set(r.index("shortName")) == set(["t", "r"])
    assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())
    assert np.allclose(ref_r_vals + 1.0, r.sel(param="r", step=6, level=500).to_numpy())

    assert sorted(r.metadata("base_datetime")) == sorted(ds_ek.metadata("base_datetime"))
    assert sorted(r.metadata("valid_datetime")) == sorted(ds_ek.metadata("valid_datetime"))


@pytest.mark.cache
def test_xr_write_seasonal():
    ds_ek = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/xr_engine/date/jma_seasonal_fc_ref_time_per_member.grib"),
    )
    ds_ek = ds_ek.sel(param="2t")
    assert len(ds_ek) == 60

    ds = ds_ek.to_xarray(
        time_dim_mode="forecast",
        dim_roles={"date": "indexingDate", "time": "indexingTime", "step": "forecastMonth"},
        dim_name_from_role_name=False,
    )

    import xarray as xr

    xr.set_options(keep_attrs=True)
    ds += 1

    # dataset
    r = ds.earthkit.to_fieldlist()
    assert len(r) == 60

    assert sorted(r.metadata(["indexingDate", "indexingTime", "forecastMonth"])) == sorted(
        ds_ek.metadata(["indexingDate", "indexingTime", "forecastMonth"])
    )


def test_xr_write_bits_per_value():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t", "r"], level=[500, 850])

    ref_bpm = ds_ek[0].metadata("bitsPerValue")
    assert ref_bpm == 16

    ds_ek = ds_ek.to_fieldlist()
    ds_ek = ds_ek.from_fields([f.clone(bitsPerValue=8) for f in ds_ek])
    assert ds_ek[0].metadata("bitsPerValue") == 8
    assert ds_ek[0].handle.get("bitsPerValue") == 0

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**{"profile": "mars", "time_dim_mode": "raw"})
    ds += 1

    # data-array
    r = ds["t"].earthkit.to_fieldlist()
    assert len(r) == 16
    assert r.index("shortName") == ["t"]
    assert r[0].metadata("bitsPerValue") == 8


@pytest.mark.cache
@pytest.mark.parametrize("method", ["to_grib", "to_target_on_obj", "to_target_func"])
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "raw"},
    ],
)
def test_xr_write_to_grib_file_dataarray(method, kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # data-array
    with temp_file() as path:
        if method == "to_grib":
            ds["t"].earthkit.to_grib(path)
        elif method == "to_target_on_obj":
            ds["t"].earthkit.to_target("file", path, encoder="grib")
        elif method == "to_target_func":
            to_target("file", path, data=ds["t"], encoder="grib")

        r = from_source("file", path)

        assert len(r) == 16
        assert r.index("shortName") == ["t"]
        assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())

        ref_base = [
            "2024-06-03T00:00:00",
            "2024-06-03T00:00:00",
            "2024-06-03T00:00:00",
            "2024-06-03T00:00:00",
            "2024-06-03T12:00:00",
            "2024-06-03T12:00:00",
            "2024-06-03T12:00:00",
            "2024-06-03T12:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T12:00:00",
            "2024-06-04T12:00:00",
            "2024-06-04T12:00:00",
            "2024-06-04T12:00:00",
        ]

        ref_valid = [
            "2024-06-03T00:00:00",
            "2024-06-03T00:00:00",
            "2024-06-03T06:00:00",
            "2024-06-03T06:00:00",
            "2024-06-03T12:00:00",
            "2024-06-03T12:00:00",
            "2024-06-03T18:00:00",
            "2024-06-03T18:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T00:00:00",
            "2024-06-04T06:00:00",
            "2024-06-04T06:00:00",
            "2024-06-04T12:00:00",
            "2024-06-04T12:00:00",
            "2024-06-04T18:00:00",
            "2024-06-04T18:00:00",
        ]

        assert r.metadata("base_datetime") == ref_base
        assert r.metadata("valid_datetime") == ref_valid


@pytest.mark.cache
@pytest.mark.parametrize("method", ["to_grib", "to_target_on_obj", "to_target_func"])
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "raw"},
    ],
)
def test_xr_write_to_grib_file_dataset(method, kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy()
    ref_r_vals = ds_ek.sel(param="r", step=6, level=500).to_numpy()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # dataset
    with temp_file() as path:
        if method == "to_grib":
            ds.earthkit.to_grib(path)
        elif method == "to_target_on_obj":
            ds.earthkit.to_target("file", path, encoder="grib")
        elif method == "to_target_func":
            to_target("file", path, data=ds, encoder="grib")

        r = from_source("file", path)
        assert len(r) == 16 * 2
        assert set(r.index("shortName")) == set(["t", "r"])
        assert np.allclose(ref_t_vals + 1.0, r.sel(param="t", step=6, level=500).to_numpy())
        assert np.allclose(ref_r_vals + 1.0, r.sel(param="r", step=6, level=500).to_numpy())

        assert sorted(r.metadata("base_datetime")) == sorted(ds_ek.metadata("base_datetime"))
        assert sorted(r.metadata("valid_datetime")) == sorted(ds_ek.metadata("valid_datetime"))


@pytest.mark.cache
@pytest.mark.parametrize("method", ["to_netcdf", "to_target_on_obj", "to_target_func"])
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "raw"},
    ],
)
def test_xr_write_to_netcdf_file_dataarray(method, kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy().flatten()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # data-array
    with temp_file() as path:
        if method == "to_netcdf":
            ds["t"].earthkit.to_netcdf(path)
        elif method == "to_target_on_obj":
            ds["t"].earthkit.to_target("file", path, encoder="netcdf")
        elif method == "to_target_func":
            to_target("file", path, data=ds["t"], encoder="netcdf")

        r = xr.open_dataset(path, engine="netcdf4", decode_times=False)

        for name in ["t"]:
            assert name in r.data_vars

        for name in ["latitude", "longitude"]:
            assert name in r.coords

        assert np.allclose(ref_t_vals + 1.0, r["t"].isel(step=1, level=0).to_numpy().flatten())


@pytest.mark.cache
@pytest.mark.parametrize("method", ["to_netcdf", "to_target_on_obj", "to_target_func"])
@pytest.mark.parametrize(
    "kwargs",
    [
        {"profile": "mars", "time_dim_mode": "raw"},
    ],
)
def test_xr_write_to_netcdf_file_dataset(method, kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek.sel(param=["t", "r"], level=[500, 850])

    ref_t_vals = ds_ek.sel(param="t", step=6, level=500).to_numpy().flatten()
    ref_r_vals = ds_ek.sel(param="r", step=6, level=500).to_numpy().flatten()

    import xarray as xr

    xr.set_options(keep_attrs=True)

    ds = ds_ek.to_xarray(**kwargs)
    ds += 1

    # dataset
    with temp_file() as path:
        if method == "to_netcdf":
            ds.earthkit.to_netcdf(path)
        elif method == "to_target_on_obj":
            ds.earthkit.to_target("file", path, encoder="netcdf")
        elif method == "to_target_func":
            to_target("file", path, data=ds, encoder="netcdf")

        r = xr.open_dataset(path, engine="netcdf4", decode_times=False)

        for name in ["t", "r"]:
            assert name in r.data_vars

        for name in ["latitude", "longitude"]:
            assert name in r.coords

        assert np.allclose(ref_t_vals + 1.0, r["t"].isel(step=1, level=0).to_numpy().flatten())
        assert np.allclose(ref_r_vals + 1.0, r["r"].isel(step=1, level=0).to_numpy().flatten())



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_remote_test_data_file


def test_bufr_concat():
    ds1 = from_source("file", earthkit_examples_file("temp_10.bufr"))
    ds2 = from_source("url", earthkit_remote_test_data_file("examples/synop_10.bufr"))
    ds = ds1 + ds2

    assert len(ds) == 20
    md = [x._header("dataCategory") for x in ds1] + [x._header("dataCategory") for x in ds2]
    assert md == [2] * 10 + [0] * 10


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import load_grib_data  # noqa: E402


class Counter:
    n = 0

    def increment(self):
        self.n += 1

    def get(self):
        return self.n


def count_calls(monkeypatch, cls, method):
    counter = Counter()

    def patched_method(self, *args, **kwargs):
        if not self.released:
            nonlocal counter
            counter.increment()
            return method(self, *args, **kwargs)

    patched_method.__name__ = method.__name__
    monkeypatch.setattr(cls, method.__name__, patched_method)
    return counter


@pytest.mark.cache
@pytest.mark.parametrize(
    "lazy_load,release_source,expected_result",
    [
        (False, False, {"call_count": 0}),
        (False, True, {"call_count": 32}),
        (True, False, {"call_count": 0}),
        (True, True, {"call_count": 0}),
    ],
)
def test_xr_engine_stream_release_source(lazy_load, release_source, expected_result, monkeypatch):
    filename = "test-data/xr_engine/level/pl_small.grib"
    ds_ek, _ = load_grib_data(filename, "url", stream=True)
    ds_ek_ref, _ = load_grib_data(filename, "url", stream=False)

    from earthkit.data.utils.xarray.fieldlist import ReleasableField

    calls_cnt = count_calls(monkeypatch, ReleasableField, ReleasableField._release)

    assert calls_cnt.get() == 0

    kwargs = {
        "lazy_load": lazy_load,
        "release_source": release_source,
    }

    ds = ds_ek.to_xarray(**kwargs)
    assert ds is not None

    assert calls_cnt.get() == expected_result["call_count"]

    # the stream is exhausted
    with pytest.raises(StopIteration):
        next(iter(ds_ek))

    assert np.allclose(
        ds_ek_ref.sel(param="t").order_by(["date", "time", "step", "levelist"]).to_numpy(),
        ds["t"].values.reshape((16, 19, 36)),
    )


@pytest.mark.cache
@pytest.mark.parametrize(
    "lazy_load,release_source,expected_result",
    [
        (False, False, {"call_count": 0, "param": "t"}),
        (False, True, {"call_count": 32, "param": None}),
        (True, False, {"call_count": 0, "param": "t"}),
        (True, True, {"call_count": 0, "param": "t"}),
    ],
)
def test_xr_engine_array_field_release_source(lazy_load, release_source, expected_result, monkeypatch):
    filename = "test-data/xr_engine/level/pl_small.grib"
    ds_ek, _ = load_grib_data(filename, "url", stream=False)
    ds_ek = ds_ek.to_fieldlist()

    from earthkit.data.utils.xarray.fieldlist import ReleasableField

    calls_cnt = count_calls(monkeypatch, ReleasableField, ReleasableField._release)

    assert calls_cnt.get() == 0

    kwargs = {
        "lazy_load": lazy_load,
        "release_source": release_source,
    }

    ds = ds_ek.to_xarray(**kwargs)
    assert ds is not None

    assert calls_cnt.get() == expected_result["call_count"]

    if not lazy_load and release_source:
        with pytest.raises(AttributeError):
            ds_ek[0].metadata("param")
    else:
        assert np.allclose(
            ds_ek.sel(param="t").order_by(["date", "time", "step", "levelist"]).to_numpy(),
            ds["t"].values.reshape((16, 19, 36)),
        )

        assert ds_ek[0].metadata("param") == expected_result["param"]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from indexing_fixtures import get_tmp_fixture  # noqa: E402


@pytest.mark.cache
def test_indexing_db_file():
    _, path = get_tmp_fixture("file")
    ds = from_source("file", path, indexing=True)
    assert hasattr(ds, "db")
    assert ds.db.count() == 18


@pytest.mark.cache
def test_indexing_db_file_multi():
    _, path = get_tmp_fixture("multi")
    ds = from_source("file", path, indexing=True)

    counts = [6, 6, 6]
    assert len(counts) == len(ds._indexes)
    for i, d in enumerate(ds._indexes):
        assert hasattr(d, "db"), f"db,{i}"
        assert d.db.count() == counts[i]


@pytest.mark.cache
def test_indexing_db_directory():
    _, path = get_tmp_fixture("directory")
    ds = from_source("file", path, indexing=True)
    assert hasattr(ds, "db")
    assert ds.db.count() == 18


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.decorators import availability
from earthkit.data.decorators import normalize


def name_no_default(name):
    return name


@pytest.mark.skip("Not implemented yet")
def test_enum_2_normalizers():
    g = name_no_default
    g = normalize("a", [1, 2])(g)
    g = normalize("a", [3, 4])(g)
    with pytest.raises(ValueError, "Duplicate normalize decorators"):
        g(a=1)


C3 = [
    {"name": "50"},
    {"name": "100"},
]


@pytest.mark.skip("Not implemented yet")
def test_enum_inconsistent_availablity_normalizers():
    g = name_no_default
    g = normalize("name", type=int)(g)
    g = availability(C3)(g)
    g(name="50")
    with pytest.raises(
        ValueError,
        match="Inconsistent types for availability and normalize for argument 'a'.",
    ):
        g(name="50")


def test_enum_no_type():
    g = normalize("name", multiple=True)(name_no_default)
    assert g(["a", "b"]) == ["a", "b"]
    assert g("a") == ["a"]


# def test_enum_cannot_find_type_2():
#     g = normalize("name", type="zzz")(name_no_default)
#     with pytest.raises(ValueError, match="Cannot infer type .*"):
#         g(["a", "b"])
#
#
# def test_enum_cannot_find_type_3():
#     g = normalize("name", type="zzz", multiple=True)(name_no_default)
#     with pytest.raises(ValueError, match="Cannot infer type .*"):
#         g(["a", "b"])


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sqlite3

import pytest

import earthkit.data
from earthkit.data.core.temporary import temp_directory


def _make_db(path):
    DATA = [
        (50, 3.3, "2001-01-01 00:00:00", 4.9),
        (51, -3, "2001-01-02 00:00:00", 7.3),
        (50.5, -1.8, "2001-01-03 00:00:00", 5.5),
    ]
    if os.path.exists(path):
        os.unlink(path)

    conn = sqlite3.connect(path)
    c = conn.cursor()
    c.execute(
        """CREATE TABLE data(
                    lat NUMBER,
                    lon NUMBER,
                    time TEXT,
                    value NUMBER)"""
    )
    c.executemany("INSERT INTO data VALUES(?,?,?,?);", DATA)
    conn.commit()


@pytest.mark.plugin
def test_demo_source_plugin():
    with temp_directory() as tmp_dir:
        path = os.path.join(tmp_dir, "test.db")

        _make_db(path)

        ds = earthkit.data.from_source(
            "demo-source",
            f"sqlite:///{path}",
            "select * from data;",
            parse_dates=["time"],
        )

        df = ds.to_pandas()
        assert len(df) == 3
        assert list(df.columns) == ["lat", "lon", "time", "value"]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1 - equivalent solutions
with open("res.grib", "wb") as f:
    ds.to_target("file", f)

with open("res.grib", "wb") as f:
    t = ekd.create_target("file", f)
    t.write(ds)

# case 2 - equivalent solutions
with open("res.grib", "wb") as f:
    for field in ds:
        field.to_target("file", f)

with open("res.grib", "wb") as f:
    t = ekd.create_target("file", f)
    for field in ds:
        t.write(field)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import numpy as np

from earthkit.data import from_source
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils.bbox import BoundingBox


def test_shapefile():
    ds = from_source("file", earthkit_test_data_file("NUTS_RG_20M_2021_3035.shp.zip"))
    assert len(ds) == 2010

    # slicing
    r = ds[1]
    assert r.NUTS_ID == "HR"
    assert r.NAME_LATN == "Hrvatska"
    assert len(r.geometry.geoms) == 16

    # iteration
    cnt = 0
    for f in ds:
        assert isinstance(f.NUTS_ID, str)
        cnt += 1

    assert cnt == 2010

    # conversion
    gpf = ds.to_pandas()
    assert len(gpf) == 2010
    r = gpf.iloc[1]
    assert r.NUTS_ID == "HR"
    assert r.NAME_LATN == "Hrvatska"
    assert len(r.geometry.geoms) == 16

    gpf = ds.to_geopandas()
    assert len(gpf) == 2010
    r = gpf.iloc[1]
    assert r.NUTS_ID == "HR"
    assert r.NAME_LATN == "Hrvatska"
    assert len(r.geometry.geoms) == 16

    v = ds.to_numpy()
    assert v.shape == (2010, 10)

    v = ds.to_numpy(flatten=True)
    assert v.shape == (2010 * 10,)


def test_shapefile_bounding_box():
    ds = from_source("file", earthkit_test_data_file("NUTS_RG_20M_2021_3035.shp.zip"))

    ref = BoundingBox(north=84.73, west=44.83, south=24.6, east=324.42)
    bb = ds.bounding_box()
    assert np.allclose(np.array(bb.as_list()), np.array(ref.as_list()))


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

from earthkit.data import FieldList
from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.sources.array_list import ArrayField
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import write_to_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import load_grib_data  # noqa: E402


# @pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("fl_type", ["file"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_clone_metadata(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    def _func1(field, key, original_metadata):
        return original_metadata[key] + 100

    def _func2(field, key, original_metadata):
        return field.mars_area

    def _func3(field, key, original_metadata):
        return original_metadata.get("param") + "_" + str(original_metadata.get("levelist"))

    def _func4(field, key, original_metadata):
        return original_metadata[key] + 200

    # ---------------
    # field
    # ---------------

    f = ds_ori[0].clone(
        param="q",
        levelist=_func1,
        mars_area=_func2,
        name=_func3,
    )

    assert f.metadata("param") == "q"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500
    assert f.metadata("levelist") == 600
    assert f.metadata("date", "param") == (20070101, "q")
    assert f.metadata("param", "date") == ("q", 20070101)
    assert np.allclose(np.array(f.metadata("mars_area")), np.array([90.0, 0.0, -90.0, 359.0]))
    assert f.metadata("name") == "t_500"

    # TODO: apply wrapped metadata to namespaces
    assert f.metadata(namespace="mars") == {
        "class": "ea",
        "date": 20070101,
        "domain": "g",
        "expver": "0001",
        "levelist": 500,
        "levtype": "pl",
        "param": "t",
        "step": 0,
        "stream": "oper",
        "time": 1200,
        "type": "an",
    }

    # write back to grib
    # we can only have ecCodes keys
    with temp_file() as tmp:
        f = ds_ori[0].clone(
            param="q",
            levelist=_func1,
        )

        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "q"
        assert f_saved.metadata("shortName") == "q"
        assert f_saved.metadata("level") == 600
        assert f_saved.metadata("levelist") == 600

    # ---------------------
    # field - repeated use
    # ---------------------

    f = ds_ori[0].clone(
        param="q",
        levelist=_func1,
        mars_area=_func2,
        name=_func3,
    )

    f = f.clone(param="pt", levelist=_func4)

    assert f.metadata("param") == "pt"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500

    # TODO: this should be 800
    assert f.metadata("levelist") == 700
    assert f.metadata("date", "param") == (20070101, "pt")
    assert f.metadata("param", "date") == ("pt", 20070101)
    assert np.allclose(np.array(f.metadata("mars_area")), np.array([90.0, 0.0, -90.0, 359.0]))
    assert f.metadata("name") == "t_500"

    # ---------------
    # fieldlist
    # ---------------

    fields = []
    for i in range(2):
        f = ds_ori[i].clone(
            param="q",
            levelist=_func1,
        )
        fields.append(f)

    ds = FieldList.from_fields(fields)

    assert ds.metadata("param") == ["q", "q"]
    assert ds.metadata("shortName") == ["t", "z"]
    assert ds.metadata("level") == [500, 500]
    assert ds.metadata("levelist") == [600, 600]

    # write back to grib
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        ds_saved = from_source("file", tmp)
        assert ds_saved.metadata("param") == ["q", "q"]
        assert ds_saved.metadata("shortName") == ["q", "q"]
        assert ds_saved.metadata("level") == [600, 600]
        assert ds_saved.metadata("levelist") == [600, 600]

    # TODO: implement the following
    # serialise
    # pickled_f = pickle.dumps(ds)
    # ds_1 = pickle.loads(pickled_f)

    # assert ds_1.metadata("param") == ["q", "q"]
    # assert ds_1.metadata("shortName") == ["q", "q"]
    # assert ds_1.metadata("level") == [600, 600]
    # assert ds_1.metadata("levelist") == [600, 600]


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_clone_values(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    vals_ori = ds_ori[0].values

    # ---------------
    # field
    # ---------------

    f = ds_ori[0].clone(values=vals_ori + 1)

    assert f.metadata("param") == "t"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500
    assert f.metadata("levelist") == 500
    assert f.metadata("date", "param") == (20070101, "t")
    assert f.metadata("param", "date") == ("t", 20070101)
    assert np.allclose(f.values, vals_ori + 1)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # write back to grib
    # we can only have ecCodes keys
    with temp_file() as tmp:
        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "t"
        assert f_saved.metadata("shortName") == "t"
        assert f_saved.metadata("level") == 500
        assert f_saved.metadata("levelist") == 500
        assert np.allclose(f_saved.values, vals_ori + 1)

    # ---------------------
    # field - repeated use
    # ---------------------

    f = ds_ori[0].clone(values=vals_ori + 1)
    f = f.clone(values=vals_ori + 2)

    assert f.metadata("param") == "t"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500
    assert f.metadata("levelist") == 500
    assert f.metadata("date", "param") == (20070101, "t")
    assert f.metadata("param", "date") == ("t", 20070101)
    assert np.allclose(f.values, vals_ori + 2)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # ---------------
    # fieldlist
    # ---------------

    fields = []
    for i in range(2):
        f = ds_ori[i].clone(values=vals_ori + i + 1)
        fields.append(f)

    ds = FieldList.from_fields(fields)

    assert ds.metadata("param") == ["t", "z"]
    assert ds.metadata("shortName") == ["t", "z"]
    assert ds.metadata("level") == [500, 500]
    assert ds.metadata("levelist") == [500, 500]
    assert np.allclose(ds[0].values, vals_ori + 1)
    assert np.allclose(ds[1].values, vals_ori + 2)

    # write back to grib
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        ds_saved = from_source("file", tmp)
        assert ds_saved.metadata("param") == ["t", "z"]
        assert ds_saved.metadata("shortName") == ["t", "z"]
        assert ds_saved.metadata("level") == [500, 500]
        assert ds_saved.metadata("levelist") == [500, 500]
        assert np.allclose(ds_saved[0].values, vals_ori + 1)
        assert np.allclose(ds_saved[1].values, vals_ori + 2)

    # TODO: implement the following
    # serialise
    # pickled_f = pickle.dumps(ds)
    # ds_1 = pickle.loads(pickled_f)

    # assert ds_1.metadata("param") == ["q", "q"]
    # assert ds_1.metadata("shortName") == ["q", "q"]
    # assert ds_1.metadata("level") == [600, 600]
    # assert ds_1.metadata("levelist") == [600, 600]


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_clone_combined(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    vals_ori = ds_ori[0].values

    def _func1(field, key, original_metadata):
        return original_metadata[key] + 100

    def _func2(field, key, original_metadata):
        return original_metadata[key] + 200

    # ---------------
    # field
    # ---------------

    f = ds_ori[0].clone(
        values=vals_ori + 1,
        param="q",
        levelist=_func1,
    )

    assert f.metadata("param") == "q"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500
    assert f.metadata("levelist") == 600
    assert f.metadata("date", "param") == (20070101, "q")
    assert f.metadata("param", "date") == ("q", 20070101)
    assert np.allclose(f.values, vals_ori + 1)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # write back to grib
    # we can only have ecCodes keys
    with temp_file() as tmp:
        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "q"
        assert f_saved.metadata("shortName") == "q"
        assert f_saved.metadata("level") == 600
        assert f_saved.metadata("levelist") == 600
        assert np.allclose(f_saved.values, vals_ori + 1)

    # ---------------------
    # field - repeated use
    # ---------------------

    f = ds_ori[0].clone(
        values=vals_ori + 1,
        param="q",
        levelist=_func1,
    )
    f = f.clone(values=vals_ori + 2, param="pt", levelist=_func2)

    assert f.metadata("param") == "pt"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500

    # this should be 800
    assert f.metadata("levelist") == 700
    assert f.metadata("date", "param") == (20070101, "pt")
    assert f.metadata("param", "date") == ("pt", 20070101)
    assert np.allclose(f.values, vals_ori + 2)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # ---------------
    # fieldlist
    # ---------------

    fields = []
    for i in range(2):
        f = ds_ori[i].clone(
            values=vals_ori + i + 1,
            param="q",
            levelist=_func1,
        )
        fields.append(f)

    ds = FieldList.from_fields(fields)

    assert ds.metadata("param") == ["q", "q"]
    assert ds.metadata("shortName") == ["t", "z"]
    assert ds.metadata("level") == [500, 500]
    assert ds.metadata("levelist") == [600, 600]
    assert np.allclose(ds[0].values, vals_ori + 1)
    assert np.allclose(ds[1].values, vals_ori + 2)

    # write back to grib
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        ds_saved = from_source("file", tmp)
        assert ds_saved.metadata("param") == ["q", "q"]
        assert ds_saved.metadata("shortName") == ["q", "q"]
        assert ds_saved.metadata("level") == [600, 600]
        assert ds_saved.metadata("levelist") == [600, 600]
        assert np.allclose(ds_saved[0].values, vals_ori + 1)
        assert np.allclose(ds_saved[1].values, vals_ori + 2)

    # TODO: implement the following
    # serialise
    # pickled_f = pickle.dumps(ds)
    # ds_1 = pickle.loads(pickled_f)

    # assert ds_1.metadata("param") == ["q", "q"]
    # assert ds_1.metadata("shortName") == ["q", "q"]
    # assert ds_1.metadata("level") == [600, 600]
    # assert ds_1.metadata("levelist") == [600, 600]


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_clone_default(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    vals_ori = ds_ori[0].values

    # ---------------
    # field
    # ---------------

    f = ds_ori[0].clone()

    assert f.metadata("param") == "t"
    assert f.metadata("shortName") == "t"
    assert f.metadata("level") == 500
    assert f.metadata("levelist") == 500
    assert f.metadata("date", "param") == (20070101, "t")
    assert f.metadata("param", "date") == ("t", 20070101)
    assert np.allclose(f.values, vals_ori)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # write back to grib
    # we can only have ecCodes keys
    with temp_file() as tmp:
        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "t"
        assert f_saved.metadata("shortName") == "t"
        assert f_saved.metadata("level") == 500
        assert f_saved.metadata("levelist") == 500
        assert np.allclose(f_saved.values, vals_ori)


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_clone_with_metadata_object(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    vals_ori = ds_ori[0].values

    md = ds_ori[0].metadata().override(shortName="q", level=600)

    f = ds_ori[0].clone(metadata=md)

    assert f.metadata("param") == "q"
    assert f.metadata("shortName") == "q"
    assert f.metadata("level") == 600
    assert f.metadata("levelist") == 600
    assert f.metadata("date", "param") == (20070101, "q")
    assert f.metadata("param", "date") == ("q", 20070101)
    assert np.allclose(f.values, vals_ori)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # write back to grib
    with temp_file() as tmp:
        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "q"
        assert f_saved.metadata("shortName") == "q"
        assert f_saved.metadata("level") == 600
        assert f_saved.metadata("levelist") == 600
        assert np.allclose(f_saved.values, vals_ori)

    with pytest.raises(ValueError):
        ds_ori[0].clone(metadata=md, param="q")


@pytest.mark.parametrize("fl_type", ["file", "array", "memory"])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_copy_to_field(fl_type, write_method):
    ds_ori, _ = load_grib_data("test4.grib", fl_type)

    vals_ori = ds_ori[0].values

    # ---------------
    # field
    # ---------------

    md = (
        ds_ori[0]
        .metadata()
        .override(
            shortName="q",
            level=600,
        )
    )

    f = ds_ori[0].copy(
        values=vals_ori + 1,
        metadata=md,
    )

    assert isinstance(f, ArrayField)
    assert f.metadata("param") == "q"
    assert f.metadata("shortName") == "q"
    assert f.metadata("level") == 600
    assert f.metadata("levelist") == 600
    assert f.metadata("date", "param") == (20070101, "q")
    assert f.metadata("param", "date") == ("q", 20070101)
    assert np.allclose(f.values, vals_ori + 1)
    assert np.allclose(ds_ori[0].values, vals_ori)

    # write back to grib
    # we can only have ecCodes keys
    with temp_file() as tmp:
        write_to_file(write_method, tmp, f)
        f_saved = from_source("file", tmp)[0]
        assert f_saved.metadata("param") == "q"
        assert f_saved.metadata("shortName") == "q"
        assert f_saved.metadata("level") == 600
        assert f_saved.metadata("levelist") == 600
        assert np.allclose(f_saved.values, vals_ori + 1)

    # ---------------
    # fieldlist
    # ---------------

    fields = []
    for i in range(2):
        md = (
            ds_ori[i]
            .metadata()
            .override(
                shortName="q",
                level=600,
            )
        )
        f = ds_ori[i].copy(
            values=vals_ori + i + 1,
            metadata=md,
        )
        assert isinstance(f, ArrayField)
        fields.append(f)

    ds = FieldList.from_fields(fields)

    assert ds.metadata("param") == ["q", "q"]
    assert ds.metadata("shortName") == ["q", "q"]
    assert ds.metadata("level") == [600, 600]
    assert ds.metadata("levelist") == [600, 600]
    assert np.allclose(ds[0].values, vals_ori + 1)
    assert np.allclose(ds[1].values, vals_ori + 2)

    # write back to grib
    with temp_file() as tmp:
        write_to_file(write_method, tmp, ds)
        ds_saved = from_source("file", tmp)
        assert ds_saved.metadata("param") == ["q", "q"]
        assert ds_saved.metadata("shortName") == ["q", "q"]
        assert ds_saved.metadata("level") == [600, 600]
        assert ds_saved.metadata("levelist") == [600, 600]
        assert np.allclose(ds_saved[0].values, vals_ori + 1)
        assert np.allclose(ds_saved[1].values, vals_ori + 2)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import mimetypes

import pytest

from earthkit.data.testing import check_unsafe_archives


@pytest.mark.skip
def test_tar_safety():
    check_unsafe_archives(".tar")


def test_tar_mimetypes():
    assert mimetypes.guess_type("x.tar") == ("application/x-tar", None)
    assert mimetypes.guess_type("x.tgz") == ("application/x-tar", "gzip")

    assert mimetypes.guess_type("x.tar.gz") == ("application/x-tar", "gzip")
    assert mimetypes.guess_type("x.tar.bz2") == ("application/x-tar", "bzip2")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
from collections import namedtuple

from earthkit.data.utils import ensure_sequence
from earthkit.data.utils.parts import PathAndParts

UrlSpecItem = namedtuple("UrlSpecItem", ["url", "parts", "kwargs"])


def parse_url_args(urls, **kwargs):
    urls = ensure_sequence(urls)
    res_url = []
    res_kwargs = []

    parts = kwargs.pop("parts", None)
    for url in urls:
        if isinstance(url, dict):
            u = url.pop("url")
            p = url.pop("parts", None)
            if p is not None:
                res_url.append([u, p])
            else:
                res_url.append(u)

            _kwarg = dict(**kwargs)
            _kwarg.update(url)
            res_kwargs.append(_kwarg)
        else:
            p = kwargs.pop("parts", None)
            if p is not None:
                parts = p

            res_url.append(url)
            res_kwargs.append(kwargs)

    # if len(res) > 1:
    #     res = sorted(res, key=lambda x: x["key"])
    return res_url, res_kwargs, parts


class UrlSourcePathAndParts(PathAndParts):
    compress = True
    sequence = True


class UrlSpec:
    def __init__(self, spec, urls_and_parts):
        """Should not be instantiated directly.
        The public API are the factory methods."""
        self.spec = spec
        self.url_and_parts = urls_and_parts

    @classmethod
    def from_urls(cls, urls, **kwargs):
        spec = []
        if isinstance(urls, UrlSpecItem):
            spec.append(urls)
            url_and_parts = UrlSourcePathAndParts.from_paths(urls.url, urls.parts)
        else:
            url, _kwargs, parts_kwarg = parse_url_args(urls, **kwargs)
            url_and_parts = UrlSourcePathAndParts.from_paths(url, parts_kwarg)
            for i, x in enumerate(url_and_parts):
                spec.append(UrlSpecItem(x[0], x[1], _kwargs[i]))
        return cls(spec, url_and_parts)

    def __len__(self):
        return len(self.spec)

    def __iter__(self):
        return iter(self.spec)

    def __getitem__(self, n):
        return self.spec[n]

    @property
    def url(self):
        return self.url_and_parts.path

    @property
    def parts(self):
        return self.url_and_parts.parts

    def zipped(self):
        return self.url_and_parts.zipped()

    def sorted(self):
        spec = sorted(copy.deepcopy(self.spec), key=lambda spec: spec[0])
        url_and_parts = self.url_and_parts.sorted()
        return UrlSpec(spec, url_and_parts)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import re

import numpy as np

from earthkit.data.wrappers import get_wrapper

ECC_SECONDS_FACTORS = {"s": 1, "m": 60, "h": 3600}
NUM_STEP_PATTERN = re.compile(r"\d+")
SUFFIX_STEP_PATTERN = re.compile(r"\d+[a-zA-Z]{1}")


def to_datetime(dt):
    if isinstance(dt, datetime.datetime):
        return dt

    if isinstance(dt, datetime.date):
        return datetime.datetime(dt.year, dt.month, dt.day)

    if hasattr(dt, "dtype") and np.issubdtype(dt.dtype, np.datetime64):
        return numpy_datetime_to_datetime(dt)

    if isinstance(dt, np.int64):
        dt = int(dt)

    dt = get_wrapper(dt)

    return to_datetime(dt.to_datetime())


def mars_like_date_list(start, end, by):
    """Return a list of datetime objects from start to end.

    Parameters
    ----------
    start : datetime.datetime
        Start datetime object
    end : datetime.datetime
        End datetime object
    by : int
        Hours between each datetime object

    Returns
    -------
    list of datetime.datetime
    """
    assert by > 0, by
    assert end >= start
    result = []
    while start <= end:
        result.append(start)
        start = start + datetime.timedelta(days=by)
    return result


def to_datetime_list(datetimes):  # noqa C901
    if isinstance(datetimes, (datetime.datetime, np.datetime64)):
        return to_datetime_list([datetimes])

    if isinstance(datetimes, (list, tuple)):
        if len(datetimes) == 3 and isinstance(datetimes[1], str) and datetimes[1].lower() == "to":
            return mars_like_date_list(to_datetime(datetimes[0]), to_datetime(datetimes[2]), 1)

        if (
            len(datetimes) == 5
            and isinstance(datetimes[1], str)
            and isinstance(datetimes[3], str)
            and datetimes[1].lower() == "to"
            and datetimes[3].lower() == "by"
        ):
            return mars_like_date_list(
                to_datetime(datetimes[0]), to_datetime(datetimes[2]), int(datetimes[4])
            )

        return [to_datetime(x) for x in datetimes]

    datetimes = get_wrapper(datetimes)

    return to_datetime_list(datetimes.to_datetime_list())


def to_date_list(obj):
    return sorted(set(to_datetime_list(obj)))


def to_time(dt):
    if isinstance(dt, float):
        dt = int(dt)

    if isinstance(dt, str):
        if len(dt) <= 4:
            dt = int(dt)
        else:
            return to_datetime(dt).time()

    if isinstance(dt, np.int64):
        dt = int(dt)

    if isinstance(dt, int):
        if dt >= 2400:
            return to_datetime(dt).time()
        else:
            h = int(dt / 100)
            m = dt % 100
            return datetime.time(hour=h, minute=m)

    if isinstance(dt, datetime.time):
        return dt

    if isinstance(dt, datetime.datetime):
        return dt.time()

    if isinstance(dt, datetime.date):
        return dt.time()

    if hasattr(dt, "dtype") and np.issubdtype(dt.dtype, np.datetime64):
        return numpy_datetime_to_datetime(dt).time()

    if hasattr(dt, "dtype") and np.issubdtype(dt.dtype, np.timedelta64):
        dt = numpy_timedelta_to_timedelta(dt)

    if isinstance(dt, datetime.timedelta):
        return datetime.time(
            hour=int(dt.total_seconds()) // 3600,
            minute=int(dt.total_seconds()) // 60 % 60,
            second=int(dt.total_seconds()) % 60,
        )

    raise ValueError(f"Failed to convert time={dt} of type={type(dt)} to datetime.time")


def to_time_list(times):
    if not isinstance(times, (list, tuple)):
        return to_time_list([times])
    return [to_time(x) for x in times]


def to_timedelta(td):
    if isinstance(td, int):
        return datetime.timedelta(hours=td)

    if isinstance(td, datetime.time):
        return datetime.timedelta(hours=td.hour, minutes=td.minute, seconds=td.second)

    # eccodes step format
    # TODO: make it work for all the ecCodes step formats
    if isinstance(td, str):
        if re.fullmatch(NUM_STEP_PATTERN, td):
            return datetime.timedelta(hours=int(td))

        if re.fullmatch(SUFFIX_STEP_PATTERN, td):
            factor = ECC_SECONDS_FACTORS.get(td[-1], None)
            if factor is None:
                raise ValueError(f"Unsupported ecCodes step units in step: {td}")
            return datetime.timedelta(seconds=int(td[:-1]) * factor)

    if isinstance(td, datetime.timedelta):
        return td

    if np.issubdtype(td, np.timedelta64):
        return numpy_timedelta_to_timedelta(td)

    raise ValueError(f"Failed to convert td={td} type={type(td)} to timedelta")


def numpy_timedelta_to_timedelta(td):
    td = td.astype("timedelta64[s]").astype(int)
    return datetime.timedelta(seconds=int(td))


def numpy_datetime_to_datetime(dt):
    dt = dt.astype("datetime64[s]").astype(int)
    return datetime.datetime.fromtimestamp(int(dt), datetime.timezone.utc).replace(tzinfo=None)


def timedeltas_to_int(td):
    def _gcd(td):
        if td.total_seconds() % 3600 == 0:
            return datetime.timedelta(hours=1)
        if td.total_seconds() % 60 == 0:
            return datetime.timedelta(minutes=1)
        if td.microseconds == 0:
            return datetime.timedelta(seconds=1)
        else:
            return td.resolution

    if not isinstance(td, (list, tuple)):
        td = [td]

    resolution = min([_gcd(x) for x in td])
    resolution_secs = int(resolution.total_seconds())
    return [int(x.total_seconds() / resolution_secs) for x in td], resolution


def date_to_grib(d):
    try:
        d = to_datetime(d)
        if isinstance(d, datetime.datetime):
            return int(d.year * 10000 + d.month * 100 + d.day)
    except Exception as e:
        raise ValueError(f"Cannot convert date={d} of type={type(d)} to grib metadata. {e}")


def time_to_grib(t):
    t = to_time(t)

    if isinstance(t, datetime.time):
        return t.hour * 100 + t.minute
    try:
        t = int(t)
        if t < 100:
            t = t * 100
    except ValueError:
        pass

    return t


def step_to_grib(step):
    if isinstance(step, (int, str)):
        return step
    elif isinstance(step, np.int64):
        return int(step)

    step = to_timedelta(step)

    if isinstance(step, datetime.timedelta):
        hours, minutes, seconds = (
            int(step.total_seconds() // 3600),
            int(step.seconds // 60 % 60),
            int(step.seconds % 60),
        )
        if seconds == 0:
            if minutes == 0:
                return hours
            else:
                return f"{hours*60}{minutes}m"
        else:
            return f"{int(step.total_seconds())}s"

    raise ValueError(f"Cannot convert step={step} of type={type(step)} to grib metadata")


def datetime_to_grib(dt):
    dt = to_datetime(dt)
    date = int(dt.strftime("%Y%m%d"))
    time = dt.hour * 100 + dt.minute
    return date, time


def datetime_from_grib(date, time):
    date = int(date)
    time = int(time)

    return datetime.datetime(
        date // 10000,
        date % 10000 // 100,
        date % 100,
        time // 100,
        time % 100,
    )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import shutil

from earthkit.data.core.temporary import temp_directory
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_file
from earthkit.data.testing import earthkit_test_data_file

TEST_GRIB_FILES = [
    earthkit_file(p)
    for p in [
        "docs/examples/test.grib",
        "docs/examples/test4.grib",
    ]
]


def unique_grib_file():
    tmp = temp_file()
    shutil.copy(earthkit_examples_file("tuv_pl.grib"), tmp.path)
    return tmp


def unique_grib_file_list():
    tmp = []
    for p in ["t", "u", "v"]:
        f = temp_file()
        shutil.copy(earthkit_test_data_file(f"{p}_pl.grib"), f.path)
        tmp.append(f)
    return tmp


def unique_grib_dir():
    def _build_dir_with_grib_files(testdir):
        os.makedirs(testdir, exist_ok=True)
        for p in ["t", "u", "v"]:
            shutil.copy(earthkit_test_data_file(f"{p}_pl.grib"), testdir)

    tmp = temp_directory()
    _build_dir_with_grib_files(tmp.path)
    return tmp


def list_of_dicts():
    prototype = {
        "gridType": "regular_ll",
        "Nx": 2,
        "Ny": 3,
        "distinctLatitudes": [-10.0, 0.0, 10.0],
        "distinctLongitudes": [0.0, 10.0],
        "_param_id": 167,
        "values": [[1, 2], [3, 4], [5, 6]],
        "date": "20180801",
        "time": "1200",
    }
    return [
        {"param": "t", "levelist": 500, **prototype},
        {"param": "t", "levelist": 850, **prototype},
        {"param": "u", "levelist": 500, **prototype},
        {"param": "u", "levelist": 850, **prototype},
        {"param": "d", "levelist": 850, **prototype},
        {"param": "d", "levelist": 600, **prototype},
    ]


def get_tmp_fixture(input_mode):
    tmp = {
        "directory": unique_grib_dir,
        "file": unique_grib_file,
        "multi": unique_grib_file_list,
    }[input_mode]()

    if isinstance(tmp, list):
        return tmp, [x.path for x in tmp]
    else:
        return tmp, tmp.path



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_HDA


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_HDA, reason="No access to WEKEO")
@pytest.mark.parametrize("prompt", [True, False])
def test_wekeo_download(prompt):
    s = from_source(
        "wekeo",
        "EO:CLMS:DAT:CLMS_GLOBAL_BA_300M_V3_MONTHLY_NETCDF",
        prompt=prompt,
        request={
            "dataset_id": "EO:CLMS:DAT:CLMS_GLOBAL_BA_300M_V3_MONTHLY_NETCDF",
            "startdate": "2019-01-01T00:00:00.000Z",
            "enddate": "2019-01-01T23:59:59.999Z",
        },
    )
    assert len(s) == 1


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
import re
from functools import lru_cache
from io import IOBase

import deprecation

from earthkit.data.decorators import normalize
from earthkit.data.decorators import normalize_grib_keys
from earthkit.data.utils.humanize import list_to_human

LOG = logging.getLogger(__name__)

NOT_IN_EDITION_1 = (
    "productDefinitionTemplateNumber",
    "typeOfGeneratingProcess",
)

CACHE = {}


class Combined:
    def __init__(self, handle, metadata):
        self.handle = handle
        self.metadata = metadata

    def __contains__(self, key):
        # return key in self.metadata or key in self.handle
        raise NotImplementedError()

    def __getitem__(self, key):
        if key in self.metadata:
            return self.metadata[key]
        return self.handle.get(key, default=None)


@deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use GribEncoder instead")
class GribCoder:
    def __init__(self, template=None, **kwargs):
        self.template = template
        self._bbox = {}
        self.kwargs = kwargs

    @normalize_grib_keys
    @normalize("date", "date")
    def _normalize_kwargs_names(self, **kwargs):
        return kwargs

    def encode(
        self,
        values,
        check_nans=False,
        metadata={},
        template=None,
        return_bytes=False,
        missing_value=9999,
        **kwargs,
    ):
        # Make a copy as we may modify it
        md = self._normalize_kwargs_names(**self.kwargs)
        md.update(self._normalize_kwargs_names(**metadata))
        md.update(self._normalize_kwargs_names(**kwargs))

        metadata = md

        compulsory = (("date", "referenceDate"), ("param", "paramId", "shortName"))

        if template is None:
            template = self.template

        if template is None:
            handle = self.handle_from_metadata(values, metadata, compulsory)
        else:
            handle = template.handle.clone()
            self.update_metadata_from_template(metadata, template, handle)

        self.update_metadata(handle, metadata, compulsory)

        if check_nans and values is not None:
            import numpy as np

            if np.isnan(values).any():
                # missing_value = np.finfo(values.dtype).max
                missing_value = missing_value
                values = np.nan_to_num(values, nan=missing_value)
                metadata["missingValue"] = missing_value
                metadata["bitmapPresent"] = 1

        if str(metadata.get("edition")) == "1":
            for k in NOT_IN_EDITION_1:
                metadata.pop(k, None)

        if int(metadata.get("deleteLocalDefinition", 0)):
            for k in ("class", "type", "stream", "expver", "setLocalDefinition"):
                metadata.pop(k, None)

        # TODO: revisit that logic
        if "generatingProcessIdentifier" not in metadata:
            metadata["generatingProcessIdentifier"] = 255
        else:
            # kee
            if metadata["generatingProcessIdentifier"] is None:
                metadata.pop("generatingProcessIdentifier")

        LOG.debug("GribOutput.metadata %s", metadata)

        single = {}
        multiple = {}
        for k, v in metadata.items():
            if isinstance(v, (int, float, str, bool)):
                single[k] = v
            else:
                multiple[k] = v

        try:
            # Try to set all metadata at once
            # This is needed when we set multiple keys that are interdependent
            handle.set_multiple(single)
        except Exception as e:
            LOG.error("Failed to set metadata at once: %s", e)
            # Try again, but one by one
            for k, v in single.items():
                handle.set(k, v)

        for k, v in multiple.items():
            handle.set(k, v)

        if values is not None:
            handle.set_values(values)

        if return_bytes:
            return handle.get_message()

        return handle

    def update_metadata(self, handle, metadata, compulsory):
        # TODO: revisit that logic
        combined = Combined(handle, metadata)

        if "step" in metadata or "endStep" in metadata:
            if combined["type"] == "an":
                metadata["type"] = "fc"

        if "time" in metadata:  # TODO, use a normalizer
            try:
                time = int(metadata["time"])
                if time < 100:
                    metadata["time"] = time * 100
            except ValueError:
                pass

        if "time" not in metadata and "date" in metadata:
            date = metadata["date"]
            metadata["time"] = date.hour * 100 + date.minute

        if "date" in metadata:
            if isinstance(metadata["date"], datetime.datetime):
                date = metadata["date"]
                metadata["date"] = date.year * 10000 + date.month * 100 + date.day
            else:
                metadata["date"] = int(metadata["date"])

        if "stream" not in metadata:
            if "number" in metadata:
                metadata["stream"] = "enfo"
                metadata.setdefault("type", "pf")

        if "number" in metadata:
            compulsory += ("numberOfForecastsInEnsemble",)
            metadata.setdefault("productDefinitionTemplateNumber", 1)  # 11 for accumulations

        if metadata.get("type") in ("pf", "cf"):
            metadata.setdefault("typeOfGeneratingProcess", 4)

        if "levelist" in metadata:
            metadata.setdefault("levtype", "pl")

        if "param" in metadata:
            param = metadata.pop("param")
            try:
                metadata["paramId"] = int(param)
            except ValueError:
                metadata["shortName"] = param

        # levtype is a readOnly key in ecCodes >= 2.33.0
        levtype_remap = {
            "pl": "isobaricInhPa",
            "ml": "hybrid",
            "pt": "theta",
            "pv": "potentialVorticity",
            "sfc": "surface",
        }
        if "levtype" in metadata:
            v = metadata.pop("levtype")
            metadata["typeOfLevel"] = levtype_remap[v]

    def handle_from_metadata(self, values, metadata, compulsory):
        from .codes import GribCodesHandle  # Lazy loading of eccodes

        if len(values.shape) == 1:
            sample = self._gg_field(values, metadata)
        elif len(values.shape) == 2:
            sample = self._ll_field(values, metadata)
        else:
            raise ValueError(f"Invalid shape {values.shape} for GRIB, must be 1 or 2 dimension ")

        # assert False, sample
        metadata.setdefault("bitsPerValue", 16)
        metadata["scanningMode"] = 0

        if "class" in metadata or "type" in metadata or "stream" in metadata or "expver" in metadata:
            # MARS labelling
            metadata["setLocalDefinition"] = 1
            # metadata['grib2LocalSectionNumber'] = 1

        for check in compulsory:
            if not isinstance(check, tuple):
                check = [check]

            if not any(c in metadata for c in check):
                choices = list_to_human([f"'{c}'" for c in check], "or")
                raise ValueError(f"Please provide a value for {choices}.")

        LOG.debug("GribCodesHandle.from_sample(%s)", sample)
        return GribCodesHandle.from_sample(sample)

    def _ll_field(self, values, metadata):
        Nj, Ni = values.shape
        metadata["Nj"] = Nj
        metadata["Ni"] = Ni

        # We assume the scanning mode north->south, west->east
        west_east = 360 / Ni

        if Nj % 2 == 0:
            north_south = 180 / Nj
            adjust = north_south / 2
        else:
            north_south = 181 / Nj
            adjust = 0

        north = 90 - adjust
        south = -90 + adjust
        west = 0
        east = 360 - west_east

        metadata["iDirectionIncrementInDegrees"] = west_east
        metadata["jDirectionIncrementInDegrees"] = north_south

        metadata["latitudeOfFirstGridPointInDegrees"] = north
        metadata["latitudeOfLastGridPointInDegrees"] = south
        metadata["longitudeOfFirstGridPointInDegrees"] = west
        metadata["longitudeOfLastGridPointInDegrees"] = east

        edition = metadata.get("edition", 2)
        levtype = metadata.get("levtype")
        if levtype is None:
            if "levelist" in metadata:
                levtype = "pl"
            else:
                levtype = "sfc"

        return f"regular_ll_{levtype}_grib{edition}"

    def _gg_field(self, values, metadata):
        GAUSSIAN = {
            6114: (32, False),
            13280: (48, False),
            24572: (64, False),
            35718: (80, False),
            40320: (96, True),
            50662: (96, False),
            88838: (128, False),
            108160: (160, True),
            138346: (160, False),
            213988: (200, False),
            348528: (256, False),
            542080: (320, False),
            843490: (400, False),
            1373624: (512, False),
            2140702: (640, False),
            5447118: (1024, False),
            6599680: (1280, True),
            8505906: (1280, False),
            20696844: (2000, False),
        }

        n = len(values)
        if n not in GAUSSIAN:
            raise ValueError(f"Unsupported GAUSSIAN grid. Number of grid points {n:,}")
        N, octahedral = GAUSSIAN[n]

        if N not in self._bbox:
            import eccodes

            self._bbox[N] = max(eccodes.codes_get_gaussian_latitudes(N))

        metadata["latitudeOfFirstGridPointInDegrees"] = self._bbox[N]
        metadata["latitudeOfLastGridPointInDegrees"] = -self._bbox[N]
        metadata["longitudeOfFirstGridPointInDegrees"] = 0

        metadata["N"] = N
        if octahedral:
            half = list(range(20, 20 + N * 4, 4))
            pl = half + list(reversed(half))
            assert len(pl) == 2 * N, (len(pl), 2 * N)
            metadata["pl"] = pl
            metadata["longitudeOfLastGridPointInDegrees"] = 360 - max(pl) / 360
            metadata["Nj"] = len(pl)
        else:
            # We just want the PL
            metadata.update(_gg_pl(N))

        edition = metadata.get("edition", 2)
        levtype = metadata.get("levtype")
        if levtype is None:
            if "levelist" in metadata:
                levtype = "pl"
            else:
                levtype = "sfc"

        if octahedral or levtype == "sfc":
            return f"reduced_gg_{levtype}_grib{edition}"
        else:
            return f"reduced_gg_{levtype}_{N}_grib{edition}"

    def update_metadata_from_template(self, metadata, template, handle):
        # the template can contain extra metadata that is not encoded in the handle
        bpv = None
        if hasattr(template, "metadata"):
            template_md = template.metadata()
            from earthkit.data.core.metadata import WrappedMetadata

            if isinstance(template_md, WrappedMetadata):
                for k in template_md.extra.keys():
                    if k != "bitsPerValue" and k not in metadata:
                        metadata[k] = template_md.get(k)

            if "bitsPerValue" not in metadata:
                bpv = template.metadata("bitsPerValue", default=None)

        # Either the handle has valid bitsPerValue or has to be extracted
        # from the template and added to the metadata to be encoded
        if "bitsPerValue" not in metadata:
            if bpv is None:
                try:
                    bpv = template.handle.get("bitsPerValue", None)
                except Exception:
                    bpv = None

            if bpv is not None and bpv > 0:
                bpv_h = handle.get("bitsPerValue", None)
                if bpv != bpv_h:
                    metadata["bitsPerValue"] = bpv


@lru_cache(maxsize=None)
def _gg_pl(N):
    import eccodes

    sample = None
    result = {}
    try:
        sample = eccodes.codes_new_from_samples(
            f"reduced_gg_pl_{N}_grib2",
            eccodes.CODES_PRODUCT_GRIB,
        )

        for key in ("N", "Ni", "Nj"):
            result[key] = eccodes.codes_get(sample, key)

        for key in (
            "latitudeOfFirstGridPointInDegrees",
            "longitudeOfFirstGridPointInDegrees",
            "latitudeOfLastGridPointInDegrees",
            "longitudeOfLastGridPointInDegrees",
            "iDirectionIncrementInDegrees",
        ):
            result[key] = eccodes.codes_get_double(sample, key)

        pl = eccodes.codes_get_long_array(sample, "pl")
        result["pl"] = pl.tolist()
        result["gridType"] = "reduced_gg"

        return result

    finally:
        if sample is not None:
            eccodes.codes_release(sample)


@deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use FileTarget instead")
class GribOutput:
    def __init__(self, file, split_output=False, template=None, **kwargs):
        self._files = {}
        self.fileobj = None
        self.filename = None

        if isinstance(file, IOBase):
            self.fileobj = file
            split_output = False
        else:
            self.filename = file

        if split_output:
            self.split_output = re.findall(r"\{(.*?)\}", self.filename)
        else:
            self.split_output = None

        self._coder = GribCoder(template=template, **kwargs)

    def close(self):
        for f in self._files.values():
            f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()

    def write(
        self,
        values,
        check_nans=False,
        metadata={},
        template=None,
        **kwargs,
    ):
        handle = self._coder.encode(
            values,
            check_nans=check_nans,
            metadata=metadata,
            template=template,
            **kwargs,
        )

        file, path = self.f(handle)
        handle.write(file)

        return handle, path

    def f(self, handle):
        if self.fileobj:
            return self.fileobj, None

        if self.split_output:
            keys = {k.split(":")[0]: handle.get(k.split(":")[0]) for k in self.split_output}
            path = self.filename.format(**keys)
        else:
            path = self.filename

        if path not in self._files:
            self._files[path] = open(path, "wb")
        return self._files[path], path


@deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use create_target() instead")
def new_grib_output(*args, **kwargs):
    return GribOutput(*args, **kwargs)


@deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use create_encoder() instead")
def new_grib_coder(*args, **kwargs):
    return GribCoder(*args, **kwargs)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

values = ds[1].values + 1

# case 1 - equivalent solutions
c = ekd.create_encoder("grib")
d = c.encode(
    values=values,
    template=ds[1],
    shortName="2t",
    generatingProcessIdentifier=255,
)
# d is a GribEncodedData

c = ekd.create_encoder("grib")
d = c.encode(ds[1], values=values, shortName="2t", generatingProcessIdentifier=255)

# case 2
c = ekd.create_encoder("grib", template=ds[1], generatingProcessIdentifier=255)
d = c.encode(
    values=values,
    shortName="2t",
)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source

NO_ADS = not os.path.exists(os.path.expanduser("~/.adsapirc"))


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_ADS, reason="No access to ADS")
@pytest.mark.parametrize("prompt", [True, False])
def test_ads_grib_1_prompt(prompt):
    s = from_source(
        "ads",
        "cams-global-reanalysis-eac4",
        variable=["particulate_matter_10um", "particulate_matter_1um"],
        area=[50, -50, 20, 50],
        date="2012-12-12",
        prompt=prompt,
        time="12:00",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_ADS, reason="No access to ADS")
def test_ads_grib_2():
    s = from_source(
        "ads",
        "cams-global-reanalysis-eac4",
        variable=["particulate_matter_10um", "particulate_matter_1um"],
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        split_on="variable",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_ADS, reason="No access to ADS")
def test_ads_grib_3():
    s = from_source(
        "ads",
        "cams-global-reanalysis-eac4",
        variable=["particulate_matter_10um", "particulate_matter_1um"],
        area=[50, -50, 20, 50],
        date="2012-12-12/to/2012-12-15",
        time="12:00",
    )
    assert len(s) == 8


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_ADS, reason="No access to ADS")
def test_ads_netcdf():
    s = from_source(
        "ads",
        "cams-global-reanalysis-eac4",
        variable=["particulate_matter_10um", "particulate_matter_1um"],
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        format="netcdf",
    )
    assert len(s) == 2


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import pickle
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)

# Testing internal structures in the xarray engine


def _pickle(data, representation):
    if representation == "file":
        with temp_file() as tmp:
            with open(tmp, "wb") as f:
                pickle.dump(data, f)

            with open(tmp, "rb") as f:
                data_res = pickle.load(f)
    elif representation == "memory":
        pickled_data = pickle.dumps(data)
        data_res = pickle.loads(pickled_data)
    else:
        raise ValueError(f"Invalid representation: {representation}")
    return data_res


@pytest.mark.cache
@pytest.mark.parametrize("representation", ["file", "memory"])
def test_xr_engine_builder_fieldlist(representation):
    ds_in = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl_small.grib"))

    from earthkit.data.utils.xarray.fieldlist import XArrayInputFieldList

    r = XArrayInputFieldList(ds_in)
    assert not isinstance(r.ds, XArrayInputFieldList)
    assert r.unwrap() is ds_in
    r_p = _pickle(r, representation)
    assert r_p is not r
    assert r_p.ds is not r.ds
    assert r_p.ds.metadata("time", astype=int) == [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
        1200,
    ]

    r0 = r.sel(param="t", level=500)
    assert not isinstance(r0.ds, XArrayInputFieldList)
    assert len(r0) == 8
    assert r0.ds.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]
    r0_uw = r0.unwrap()
    assert not isinstance(r0_uw, XArrayInputFieldList)
    assert r0_uw.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]
    r0_p = _pickle(r0, representation)
    assert r0_p is not r0
    assert r0_p.ds is not r0.ds
    assert r0_p.ds.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]

    r1 = r0.order_by("time")
    assert not isinstance(r1.ds, XArrayInputFieldList)
    assert r1.ds.metadata("time", astype=int) == [0, 0, 0, 0, 1200, 1200, 1200, 1200]
    r1_uw = r1.unwrap()
    assert not isinstance(r1_uw, XArrayInputFieldList)
    assert r1_uw.metadata("time", astype=int) == [0, 0, 0, 0, 1200, 1200, 1200, 1200]
    r1_p = _pickle(r1, representation)
    assert r1_p is not r1
    assert r1_p.ds is not r1.ds
    assert r1_p.ds.metadata("time", astype=int) == [0, 0, 0, 0, 1200, 1200, 1200, 1200]

    r2 = r1.order_by("step")
    assert not isinstance(r2.ds, XArrayInputFieldList)
    assert r2.ds.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]
    assert r2.ds.metadata("step", astype=int) == [0, 0, 0, 0, 6, 6, 6, 6]
    r2_uw = r2.unwrap()
    assert not isinstance(r2_uw, XArrayInputFieldList)
    assert r2_uw.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]
    assert r2_uw.metadata("step", astype=int) == [0, 0, 0, 0, 6, 6, 6, 6]
    r2_p = _pickle(r2, representation)
    assert r2_p is not r2
    assert r2_p.ds is not r2.ds
    assert r2_p.ds.metadata("time", astype=int) == [0, 0, 1200, 1200, 0, 0, 1200, 1200]
    assert r2_p.ds.metadata("step", astype=int) == [0, 0, 0, 0, 6, 6, 6, 6]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import shutil

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import make_tgz
from earthkit.data.testing import make_zip
from earthkit.data.testing import preserve_cwd
from earthkit.data.testing import write_to_file

LOG = logging.getLogger(__name__)


def test_file_source_grib():
    from earthkit.data.readers.grib.file import GRIBReader

    s = from_source("file", earthkit_examples_file("test.grib"))

    assert isinstance(s, GRIBReader), s
    assert len(s) == 2


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_source_grib_save(write_method):
    ds = from_source("file", os.path.abspath(earthkit_examples_file("test.grib")))
    with temp_directory() as tmpdir:
        # Check file save to assigned filename
        f_tmp = os.path.join(tmpdir, "test2.grib")
        write_to_file(write_method, f_tmp, ds)
        assert os.path.isfile(f_tmp)
        # Check file can be saved in current dir with detected filename:
        with preserve_cwd():
            os.chdir(tmpdir)
            write_to_file(write_method, f_tmp, ds)
            ds.save()
            assert os.path.isfile("test.grib")


@pytest.mark.parametrize("write_method", ["save", "target"])
def test_file_source_grib_no_overwrite(write_method):
    ds = from_source("file", os.path.abspath(earthkit_examples_file("test.grib")))
    with temp_directory() as tmpdir:
        with preserve_cwd():
            os.chdir(tmpdir)
            # Save the file locally
            write_to_file(write_method, "test.grib", ds)
            # Open the local file
            ds1 = from_source("file", "test.grib")
            with pytest.warns(
                UserWarning,
                match="Earthkit refusing to overwrite",
            ):
                write_to_file(write_method, "test.grib", ds1)
            with pytest.warns(
                UserWarning,
                match="Earthkit refusing to overwrite",
            ):
                ds1.save()


def test_file_source_netcdf():
    s = from_source("file", earthkit_examples_file("test.nc"))
    assert len(s) == 2


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_source_netcdf_save(write_method):
    ds = from_source("file", os.path.abspath(earthkit_examples_file("test.nc")))
    with temp_directory() as tmpdir:
        # Check file save to assigned filename
        f_tmp = os.path.join(tmpdir, "test2.nc")
        write_to_file(write_method, f_tmp, ds)
        assert os.path.isfile(f_tmp)
        # Check file can be saved in current dir with detected filename:
        with preserve_cwd():
            os.chdir(tmpdir)
            ds.save()
            assert os.path.isfile("test.nc")


@pytest.mark.parametrize("write_method", ["save", "target"])
def test_file_source_netcdf_no_overwrite(write_method):
    ds = from_source("file", os.path.abspath(earthkit_examples_file("test.nc")))
    with temp_directory() as tmpdir:
        with preserve_cwd():
            os.chdir(tmpdir)
            # Save the file locally
            write_to_file(write_method, "test.nc", ds)
            # Open the local file
            ds1 = from_source("file", "test.nc")
            with pytest.warns(
                UserWarning,
                match="Earthkit refusing to overwrite",
            ):
                write_to_file(write_method, "test.nc", ds1)
            with pytest.warns(
                UserWarning,
                match="Earthkit refusing to overwrite",
            ):
                ds1.save()


def test_file_source_odb():
    s = from_source("file", earthkit_examples_file("test.odb"))
    assert s.path == earthkit_examples_file("test.odb")


# def test_user_1():
#     s = from_source("file", earthkit_file("docs/examples/test.grib"))
#     home_file = os.path.expanduser("~/.earthkit_data/test.grib")
#     try:
#         s.save(home_file)
#         # Test expand user
#         s = from_source("file", "~/.earthkit_data/test.grib")
#         assert len(s) == 2
#     finally:
#         try:
#             os.unlink(home_file)
#         except OSError:
#             LOG.exception("unlink(%s)", home_file)


# @pytest.mark.skipif(
#     sys.platform == "win32", reason="Cannot (yet) use expandvars on Windows"
# )
# def test_user_2():
#     s = from_source("file", earthkit_file("docs/examples/test.grib"))
#     home_file = os.path.expanduser("~/.earthkit_data/test.grib")
#     try:
#         s.save(home_file)
#         # Test expand vars
#         s = from_source("file", "$HOME/.earthkit_data/test.grib", expand_vars=True)
#         assert len(s) == 2
#     finally:
#         try:
#             os.unlink(home_file)
#         except OSError:
#             LOG.exception("unlink(%s)", home_file)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_glob(write_method):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), ds)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), ds)

        ds = from_source("file", os.path.join(tmpdir, "*.grib"))
        assert len(ds) == 4, len(ds)

        ds = from_source("file", tmpdir)
        assert len(ds) == 4, len(ds)


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_single_directory(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), s2)

        ds = from_source("file", tmpdir)
        assert len(ds) == 6, len(ds)

        ref = [
            ("2t", 0),
            ("msl", 0),
            ("t", 500),
            ("z", 500),
            ("t", 850),
            ("z", 850),
        ]
        assert ds.metadata(("param", "level")) == ref


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_multi_directory(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            ds = from_source("file", [tmpdir1, tmpdir2])
            assert len(ds) == 14, len(ds)

            ref = [
                ("2t", 0),
                ("msl", 0),
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            assert ds.metadata(("param", "level")) == ref


@pytest.mark.parametrize("filter_kwarg", [(lambda x: "b.grib" in x), ("*b.grib")])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_single_directory_filter(filter_kwarg, write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    with temp_directory() as tmpdir:
        write_to_file(write_method, os.path.join(tmpdir, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir, "b.grib"), s2)

        ds = from_source("file", tmpdir, filter=filter_kwarg)
        assert len(ds) == 4, len(ds)

        ref = [
            ("t", 500),
            ("z", 500),
            ("t", 850),
            ("z", 850),
        ]
        assert ds.metadata(("param", "level")) == ref


@pytest.mark.parametrize("filter_kwarg", [(lambda x: "b.grib" in x), ("*b.grib")])
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_multi_directory_filter(filter_kwarg, write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            ds = from_source("file", [tmpdir1, tmpdir2], filter=filter_kwarg)
            assert len(ds) == 10, len(ds)

            ref = [
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            assert ds.metadata(("param", "level")) == ref


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_file_multi_directory_with_tar(write_method):
    s1 = from_source("file", earthkit_examples_file("test.grib"))
    s2 = from_source("file", earthkit_examples_file("test4.grib"))
    s3 = from_source("file", earthkit_examples_file("test6.grib"))
    with temp_directory() as tmpdir1:
        write_to_file(write_method, os.path.join(tmpdir1, "a.grib"), s1)
        write_to_file(write_method, os.path.join(tmpdir1, "b.grib"), s2)

        with temp_directory() as tmpdir2:
            write_to_file(write_method, os.path.join(tmpdir2, "a.grib"), s1)
            write_to_file(write_method, os.path.join(tmpdir2, "b.grib"), s3)

            paths = [os.path.join(tmpdir2, f) for f in ["a.grib", "b.grib"]]
            make_tgz(tmpdir2, "test.tar.gz", paths)

            ds = from_source("file", [tmpdir1, tmpdir2])
            assert len(ds) == 22, len(ds)

            ref = [
                ("2t", 0),
                ("msl", 0),
                ("t", 500),
                ("z", 500),
                ("t", 850),
                ("z", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
                ("2t", 0),
                ("msl", 0),
                ("t", 1000),
                ("u", 1000),
                ("v", 1000),
                ("t", 850),
                ("u", 850),
                ("v", 850),
            ]

            assert ds.metadata(("param", "level")) == ref


def test_file_grib_tar_with_single_file():
    with temp_directory() as tmpdir:
        path = os.path.join(tmpdir, "a.grib")
        shutil.copy(earthkit_examples_file("test.grib"), path)
        make_tgz(tmpdir, "test.tar.gz", [path])
        t_path = os.path.join(tmpdir, "test.tar.gz")

        ds = from_source("file", t_path)
        assert len(ds) == 2, len(ds)
        assert os.path.basename(ds.path) == "a.grib"


def test_file_grib_zip_with_single_file():
    with temp_directory() as tmpdir:
        path = os.path.join(tmpdir, "a.grib")
        shutil.copy(earthkit_examples_file("test.grib"), path)
        make_zip(tmpdir, "test.zip", [path])
        t_path = os.path.join(tmpdir, "test.zip")

        ds = from_source("file", t_path)
        assert len(ds) == 2, len(ds)
        assert os.path.basename(ds.path) == "a.grib"


def test_file_netcdf_tar_with_single_file():
    with temp_directory() as tmpdir:
        path = os.path.join(tmpdir, "a.nc")
        shutil.copy(earthkit_examples_file("test.nc"), path)
        make_tgz(tmpdir, "test.tar.gz", [path])
        t_path = os.path.join(tmpdir, "test.tar.gz")

        ds = from_source("file", t_path)
        assert len(ds) == 2, len(ds)
        assert os.path.basename(ds.path) == "a.nc"


def test_file_netcdf_zip_with_single_file_1():
    with temp_directory() as tmpdir:
        path = os.path.join(tmpdir, "a.nc")
        shutil.copy(earthkit_examples_file("test.nc"), path)
        make_zip(tmpdir, "test.zip", [path])
        ar_path = os.path.join(tmpdir, "test.zip")

        ds = from_source("file", ar_path)
        assert len(ds) == 2, len(ds)
        assert os.path.basename(ds.path) == "a.nc"


def test_file_netcdf_zip_with_single_file_2():
    # This NetCDF is not read into a fieldlist
    import xarray as xr

    N = 3
    ds_in = xr.DataArray(range(N), coords={"dim_0": range(N)}, name="foo").to_dataset()
    # print(ds_in)

    with temp_directory() as tmpdir:

        path = os.path.join(tmpdir, "a.nc")
        ds_in["foo"].to_netcdf(path)

        make_zip(tmpdir, "test.zip", [path])
        z_path = os.path.join(tmpdir, "test.zip")

        ds = from_source("file", z_path)

        # print(f"{ds=}")
        assert os.path.basename(ds.path) == "a.nc"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math
import os
import threading
from abc import abstractmethod
from collections import defaultdict

from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.index import MaskIndex
from earthkit.data.core.index import MultiIndex
from earthkit.data.decorators import alias_argument
from earthkit.data.decorators import detect_out_filename
from earthkit.data.indexing.database import FILEPARTS_KEY_NAMES
from earthkit.data.indexing.database import MORE_KEY_NAMES
from earthkit.data.indexing.database import MORE_KEY_NAMES_WITH_UNDERSCORE
from earthkit.data.indexing.database import STATISTICS_KEY_NAMES
from earthkit.data.readers.grib.codes import GribField
from earthkit.data.readers.grib.pandas import PandasMixIn
from earthkit.data.readers.grib.xarray import XarrayMixIn
from earthkit.data.utils.availability import Availability

# from earthkit.data.utils.progbar import progress_bar

LOG = logging.getLogger(__name__)


class GribFieldList(PandasMixIn, XarrayMixIn, FieldList):
    r"""Represents a list of :obj:`GribField <data.readers.grib.codes.GribField>`\ s.

    We can **iterate** through the fields as follows:

    >>> import earthkit.data
    >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
    >>> len(ds)
    6
    >>> for f in ds:
    ...     print(f)
    ...
    GribField(t,1000,20180801,1200,0,0)
    GribField(u,1000,20180801,1200,0,0)
    GribField(v,1000,20180801,1200,0,0)
    GribField(t,850,20180801,1200,0,0)
    GribField(u,850,20180801,1200,0,0)
    GribField(v,850,20180801,1200,0,0)

    :obj:`Fieldset` objects can be **concatenated** with the + operator:

    >>> import earthkit.data
    >>> ds1 = earthkit.data.from_source("file", "docs/examples/test.grib")
    >>> len(ds1)
    2
    >>> ds2 = earthkit.data.from_source("file", "docs/examples/test6.grib")
    >>> len(ds2)
    6
    >>> ds = ds1 + ds2
    >>> len(ds)
    8

    Standard Python slicing works:

    >>> import earthkit.data
    >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
    >>> len(ds)
    6
    >>> ds[0]
    GribField(t,1000,20180801,1200,0,0)
    >>> for f in ds[0:3]:
    ...     print(f)
    GribField(t,1000,20180801,1200,0,0)
    GribField(u,1000,20180801,1200,0,0)
    GribField(v,1000,20180801,1200,0,0)
    >>> for f in ds[0:4:2]:
    ...     print(f)
    GribField(t,1000,20180801,1200,0,0)
    GribField(v,1000,20180801,1200,0,0)
    >>> ds[-1]
    GribField(v,850,20180801,1200,0,0)
    >>> for f in ds[-2:]:
    ...     print(f)
    GribField(u,850,20180801,1200,0,0)
    GribField(v,850,20180801,1200,0,0)

    Slicing also works with a list or an ndarray:

    >>> for f in ds[[1, 3]]:
    ...     print(f)
    ...
    GribField(u,1000,20180801,1200,0,0)
    GribField(t,850,20180801,1200,0,0)
    >>> for f in ds[np.array([1, 3])]:
    ...     print(f)
    ...
    GribField(u,1000,20180801,1200,0,0)
    GribField(t,850,20180801,1200,0,0)

    """

    _availability = None

    def __init__(self, *args, **kwargs):
        if self.availability_path is not None and os.path.exists(self.availability_path):
            self._availability = Availability(self.availability_path)

        # Index.__init__(self, *args, **kwargs)
        FieldList.__init__(self, *args, **kwargs)

    @classmethod
    def new_mask_index(cls, *args, **kwargs):
        return GribMaskFieldList(*args, **kwargs)

    @property
    def availability_path(self):
        return None

    @classmethod
    def merge(cls, sources):
        if not all(isinstance(_, GribFieldList) for _ in sources):
            raise ValueError("GribFieldList can only be merged to another GribFieldLists")
        return GribMultiFieldList(sources)

    def _custom_availability(self, ignore_keys=None, filter_keys=lambda k: True):
        from earthkit.data.utils.progbar import progress_bar

        def dicts():
            for i in progress_bar(iterable=range(len(self)), desc="Building availability"):
                dic = self.get_metadata(i)

                for k in list(dic.keys()):
                    if not filter_keys(k):
                        dic.pop(k)
                        continue
                    if ignore_keys and k in ignore_keys:
                        dic.pop(k)
                        continue
                    if dic[k] is None:
                        dic.pop(k)
                        continue

                yield dic

        from earthkit.data.utils.availability import Availability

        return Availability(dicts())

    @property
    def availability(self):
        if self._availability is not None:
            return self._availability
        LOG.debug("Building availability")

        self._availability = self._custom_availability(
            ignore_keys=FILEPARTS_KEY_NAMES
            + STATISTICS_KEY_NAMES
            + MORE_KEY_NAMES_WITH_UNDERSCORE
            + MORE_KEY_NAMES
        )
        return self.availability

    def _is_full_hypercube(self):
        non_empty_coords = {k: v for k, v in self.availability._tree.unique_values().items() if len(v) > 1}
        expected_size = math.prod([len(v) for k, v in non_empty_coords.items()])
        return len(self) == expected_size

    @alias_argument("levelist", ["level", "levellist"])
    @alias_argument("levtype", ["leveltype"])
    @alias_argument("param", ["variable", "parameter"])
    @alias_argument("number", ["realization", "realisation"])
    @alias_argument("class", "klass")
    def _normalize_kwargs_names(self, **kwargs):
        return kwargs

    @detect_out_filename
    def save(self, filename, append=False, bits_per_value=None):
        r"""Write all the fields into a file.

        Parameters
        ----------
        filename: str
            The target file path.
        append: bool
            When it is true append data to the target file. Otherwise
            the target file be overwritten if already exists.
        bits_per_value: int or None
            Set the ``bitsPerValue`` GRIB key for each message in the generated
            output. When None the ``bitsPerValue`` stored in the message metadata
            will be used.

        See Also
        --------
        write

        """
        super().save(
            filename,
            append=append,
            bits_per_value=bits_per_value,
        )


class GribMaskFieldList(GribFieldList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_mask(self, self)

    def __getstate__(self):
        r = {}
        r["mask_index"] = self._index
        r["mask_indices"] = self._indices
        return r

    def __setstate__(self, state):
        _index = state["mask_index"]
        _indices = state["mask_indices"]
        self.__init__(_index, _indices)


class GribMultiFieldList(GribFieldList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)
        FieldList._init_from_multi(self, self)

    def __getstate__(self):
        r = {}
        r["multi_indexes"] = self._indexes
        r["kwargs"] = self._kwargs
        return r

    def __setstate__(self, state):
        self.__init__(
            state["multi_indexes"],
            **state["kwargs"],
        )


class GribFieldManager:
    def __init__(self, policy, owner):
        self.policy = policy
        self.cache = None
        self.lock = threading.Lock()

        if self.policy == "persistent":
            from lru import LRU

            # TODO: the number of fields might only be available only later (e.g. fieldlists with
            # an SQL index). Consider making cache a cached property.
            n = len(owner)
            if n > 0:
                self.cache = LRU(n)

        self.field_create_count = 0

        # check consistency
        if self.cache is not None:
            assert self.policy == "persistent"

    def field(self, n, create):
        if self.cache is not None:
            with self.lock:
                if n in self.cache:
                    return self.cache[n]
                else:
                    field = create(n)
                    self._field_created()
                    self.cache[n] = field
                    return field
        else:
            self._field_created()
            return create(n)

    def _field_created(self):
        self.field_create_count += 1

    def diag(self):
        r = defaultdict(int)
        r["grib_field_policy"] = self.policy
        if self.cache is not None:
            r["field_cache_size"] = len(self.cache)

        r["field_create_count"] = self.field_create_count
        return r


class GribHandleManager:
    # TODO: split into policies
    def __init__(self, policy, cache_size):
        self.policy = policy
        self.max_cache_size = cache_size
        self.cache = None
        self.lock = threading.Lock()

        if self.policy == "cache":
            if self.max_cache_size > 0:
                from lru import LRU

                self.cache = LRU(self.max_cache_size)
            else:
                raise ValueError(
                    'grib_handle_cache_size must be greater than 0 when grib_handle_policy="cache"'
                )

        self.handle_create_count = 0

        # check consistency
        if self.cache is not None:
            self.policy == "cache"
        else:
            self.policy in ["persistent", "temporary"]

    def handle(self, field, create):
        if self.policy == "cache":
            key = (field.path, field._offset)
            with self.lock:
                if key in self.cache:
                    return self.cache[key]
                else:
                    handle = create()
                    self._handle_created()
                    self.cache[key] = handle
                    return handle
        elif self.policy == "persistent":
            if field._handle is None:
                with self.lock:
                    if field._handle is None:
                        field._handle = create()
                        self._handle_created()
                    return field._handle
            return field._handle
        elif self.policy == "temporary":
            self._handle_created()
            return create()

    def _handle_created(self):
        self.handle_create_count += 1

    def diag(self):
        r = defaultdict(int)
        r["grib_handle_policy"] = self.policy
        r["grib_handle_cache_size"] = self.max_cache_size
        if self.cache is not None:
            r["handle_cache_size"] = len(self.cache)

        r["handle_create_count"] = self.handle_create_count
        return r


class GribFieldListInFiles(GribFieldList):
    def __init__(
        self,
        *args,
        grib_field_policy=None,
        grib_handle_policy=None,
        grib_handle_cache_size=None,
        use_grib_metadata_cache=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        from earthkit.data.core.config import CONFIG

        def _get_opt(v, name):
            return v if v is not None else CONFIG.get(name)

        self._field_manager = GribFieldManager(_get_opt(grib_field_policy, "grib-field-policy"), self)
        self._handle_manager = GribHandleManager(
            _get_opt(grib_handle_policy, "grib-handle-policy"),
            _get_opt(grib_handle_cache_size, "grib-handle-cache-size"),
        )

        self._use_metadata_cache = _get_opt(use_grib_metadata_cache, "use-grib-metadata-cache")

    def _create_field(self, n):
        part = self.part(n)
        field = GribField(
            part.path,
            part.offset,
            part.length,
            handle_manager=self._handle_manager,
            use_metadata_cache=self._use_metadata_cache,
        )
        if field is None:
            raise RuntimeError(f"Could not get a handle for part={part}")
        return field

    def _getitem(self, n):
        if isinstance(n, int):
            if n < 0:
                n += len(self)
            if n >= len(self):
                raise IndexError(f"Index {n} out of range")

            return self._field_manager.field(n, self._create_field)

    def __len__(self):
        return self.number_of_parts()

    def _cache_diag(self):
        """For testing only"""
        r = defaultdict(int)
        r.update(self._field_manager.diag())
        r.update(self._handle_manager.diag())

        if self._field_manager.cache is not None:
            from earthkit.data.utils.diag import collect_field_metadata_cache_diag

            for f in self._field_manager.cache.values():
                if f._handle is not None:
                    r["current_handle_count"] += 1

                if self._use_metadata_cache:
                    collect_field_metadata_cache_diag(f, r)
        return r

    @abstractmethod
    def part(self, n):
        self._not_implemented()

    @abstractmethod
    def number_of_parts(self):
        self._not_implemented()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from abc import ABCMeta
from abc import abstractmethod
from collections import defaultdict
from functools import cached_property

from earthkit.data.utils import ensure_dict
from earthkit.data.utils import ensure_iterable

LOG = logging.getLogger(__name__)


class CFAttrs:
    def _load(self):
        here = os.path.dirname(__file__)
        path = os.path.join(here, "cf_attrs.yaml")
        if os.path.exists(path):
            import yaml

            try:
                with open(path, "r") as f:
                    return yaml.safe_load(f)
            except Exception as e:
                LOG.exception(f"Failed read CF attributes file {path}. {e}")
                raise
        else:
            raise ValueError(f"CF attributes file not found! path={path}")

    @cached_property
    def attrs(self):
        return self._load()

    def can_be_global(self, name):
        item = self.attrs.get(name, None)
        if item:
            return "G" in item["use"]
        return True


CF_ATTRS = CFAttrs()


class Attr:
    """Generic attribute class.

    Parameters
    ----------
    name : str
        The name of the attribute.
    """

    def __init__(self, name):
        self.name = name

    def value(self):
        return None

    def fixed(self):
        False


class KeyAttr(Attr):
    """Metadata key attribute class.

    Parameters
    ----------
    name : str
        The name of the attribute.
    key : str, None
        The metadata key to retrieve the attribute value. If None,
        the name is used as the key.
    """

    def __init__(self, name, key=None):
        super().__init__(name)
        self.key = key if key is not None else name

    def get(self, metadata):
        return {self.name: metadata.get(self.key, default=None)}

    def __repr__(self) -> str:
        return f"KeyAttr({self.name})"


class NamespaceAttr(Attr):
    """Namespace attribute.

    Parameters
    ----------
    name : str
        The name of the metadata namespace
    """

    def __init__(self, name, ns=None):
        super().__init__(name)
        self.ns = ns if ns is not None else name

    def get(self, metadata):
        return {self.name: metadata.as_namespace(self.ns)}

    def __repr__(self) -> str:
        return f"NamespaceAttr({self.name}, ns={self.ns})"


class FixedAttr(Attr):
    """Fixed attribute.

    Parameters
    ----------
    name : str
        The name of the attribute.
    value : str, bool, int, float
        The value of the attribute.
    """

    def __init__(self, name, value):
        super().__init__(name)
        self._value = value

    def value(self):
        return self._value

    def get(self, metadata):
        return self._value

    def fixed(self):
        return True

    def __repr__(self) -> str:
        return f"FixedAttr({self.name}, {self._value})"


class CallableAttr(Attr):
    """Callable attribute.

    Parameters
    ----------
    name : str, None
        The name of the attribute. Can be None.
    func : callable
        The function that returns the value of the attribute.
    """

    def __init__(self, name, func):
        super().__init__(name)
        self.func = func

    def __call__(self, metadata):
        if self.name is None:
            return self.func(metadata)
        else:
            res = self.func(self.name, metadata)
            return {self.name: res}

    def __repr__(self) -> str:
        return f"CallableAttr({self.name})"


class AttrList(list):
    def __init__(self, attrs):
        attrs = ensure_iterable(attrs)

        for k in attrs:
            if isinstance(k, dict):
                for k1, v1 in k.items():
                    self.append(self._make(v1, name=k1))
            else:
                self.append(self._make(k))

    @staticmethod
    def _make(val, name=None):
        if name is not None and not isinstance(name, str):
            raise ValueError(f"Attribute name must be a string: {name=}")

        if isinstance(val, Attr):
            return val
        elif isinstance(val, str):
            if val.startswith("namespace="):
                ns = val[len("namespace=") :]
                if name is None:
                    name = ns
                return NamespaceAttr(name, ns=ns)
            elif val.startswith("key="):
                key = val[len("key=") :]
                if name is None:
                    name = key
                return KeyAttr(name, key=key)
            elif name is None:
                name = val
                key = val
                return KeyAttr(val, key=key)
            else:
                return FixedAttr(name, val)
        elif callable(val):
            return CallableAttr(name, val)
        elif name is not None:
            return FixedAttr(name, val)
        else:
            raise ValueError(f"Invalid attribute: {name=} value={val}")

    def fixed(self):
        return [attr for attr in self if attr.fixed()]

    def non_fixed(self):
        return [attr for attr in self if not attr.fixed()]


class Attrs:
    def __init__(self, profile, attrs_mode, attrs, variable_attrs, global_attrs, coord_attrs, rename_attrs):
        self.profile = profile
        self.attrs = AttrList(attrs)
        self.variable_attrs = AttrList(variable_attrs)
        self.global_attrs = AttrList(global_attrs)
        self.coord_attrs = ensure_dict(coord_attrs)
        self.rename_attrs_map = ensure_dict(rename_attrs)

        # attrs must be unique and contain all variable and global attrs
        for k in self.variable_attrs + self.global_attrs:
            if k not in self.attrs:
                self.attrs.append(k)

        self.builder = self._make_builder(attrs_mode)(profile, self)

    @staticmethod
    def _make_builder(mode):
        if mode == "unique":
            return UniqueAttrBuilder
        elif mode == "fixed":
            return FixedAttrBuilder
        else:
            raise ValueError(f"Invalid attrs_mode: {mode}. Must be one of 'unique', 'fixed'.")


class AttrsBuilder(metaclass=ABCMeta):
    def __init__(self, profile, attrs):
        self.profile = profile
        self.attrs = attrs

    @abstractmethod
    def _build(self, *args, **kwargs):
        pass

    def build(self, *args, rename=True, **kwargs):
        def _rename(d):
            return {self.attrs.rename_attrs_map.get(k, k): v for k, v in d.items()}

        def _id(x):
            return x

        if rename:
            rename = _rename
        else:
            rename = _id

        global_attrs = rename(self._build(*args, rename=rename, **kwargs))

        # add fixed global attrs
        for item in self.attrs.global_attrs.fixed():
            if item.name not in global_attrs:
                global_attrs[item.name] = item.value()

        # TODO: make it optional
        # global_attrs.pop("units", None)

        return global_attrs


class UniqueAttrBuilder(AttrsBuilder):
    def _build(self, ds, t_vars, rename=None):
        attrs = defaultdict(set)

        for var_obj in t_vars.values():
            s, _ = var_obj.collect_attrs(self.attrs.attrs, strict=self.profile.strict)
            for k, v in s.items():
                attrs[k].update(ensure_iterable(v))

        global_attrs = defaultdict(list)
        for k, v in attrs.items():
            if len(v) == 1 and k not in self.attrs.variable_attrs:
                global_attrs[k] = list(v)[0]

        # Some attrs cannot be global according to the CF convention.
        # These are removed from global attrs and kept as variable attrs.
        global_attrs_keys = list(global_attrs.keys())
        global_attrs_renamed_keys = global_attrs_keys
        if rename:
            global_attrs_renamed_keys = list(rename(global_attrs).keys())

        for k1, k2 in zip(global_attrs_keys, global_attrs_renamed_keys):
            if not CF_ATTRS.can_be_global(k1) or not CF_ATTRS.can_be_global(k2):
                global_attrs.pop(k1)

        for k in self.attrs.variable_attrs:
            if k in global_attrs:
                global_attrs.pop(k)

        for var_obj in t_vars.values():
            var_obj.adjust_attrs(drop_keys=global_attrs.keys(), rename=rename)

        global_attrs = {k: v for k, v in global_attrs.items() if v is not None}

        return global_attrs


class FixedAttrBuilder(AttrsBuilder):
    def _build(self, ds, t_vars, rename=None):
        global_attrs = dict()

        for i, var_obj in enumerate(t_vars.values()):
            if i == 0:
                _, global_attrs = var_obj.collect_attrs(
                    self.attrs.variable_attrs,
                    strict=self.profile.strict,
                    extra_attrs=self.attrs.global_attrs.non_fixed(),
                )
            else:
                var_obj.collect_attrs(self.attrs.variable_attrs, strict=self.profile.strict)

            var_obj.adjust_attrs(rename=rename)

        return global_attrs



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from . import Reader

LOG = logging.getLogger(__name__)


class UnknownReaderBase(Reader):
    def __init__(
        self,
        source,
        path="",
        magic=None,
        content_type=None,
        skip_warning=False,
        **kwargs,
    ):
        super().__init__(source, path)
        self.magic = magic
        self.content_type = content_type
        self.skip_warning = skip_warning

    def ignore(self):
        # Used by multi-source
        return True

    def __len__(self):
        return 0


class UnknownReader(UnknownReaderBase):
    def __init__(self, source, path, **kwargs):
        super().__init__(source, path=path, **kwargs)
        if not self.skip_warning:
            LOG.warning(
                (
                    f"Unknown file type, no reader available. "
                    f"path={path} magic={self.magic} content_type={self.content_type}"
                )
            )


class UnknownStreamReader(UnknownReaderBase):
    def __init__(self, source, data, **kwargs):
        super().__init__(source, **kwargs)
        if not self.skip_warning:
            LOG.warning(
                (
                    f"Unknown stream data type, no reader available. "
                    f"magic={self.magic} content_type={self.content_type}"
                )
            )


class UnknownMemoryReader(UnknownReaderBase):
    def __init__(self, source, data, **kwargs):
        super().__init__(source, **kwargs)
        if not self.skip_warning:
            LOG.warning(
                (
                    f"Unknown memory data type, no reader available. "
                    f"magic={self.magic} content_type={self.content_type}"
                )
            )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import mimetypes

import pytest

from earthkit.data import from_source


def test_csv_1():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        lines=[
            [1, 2, 3],
            [4, 5, 6],
            [7, 8, 9],
        ],
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert list(df.columns) == ["a", "b", "c"]

    ds = s.to_xarray()
    assert len(ds) == 3
    assert set(ds.variables) == set(["index", "a", "b", "c"])


def test_csv_2():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        lines=[
            [1, None, 3],
            [4, 5, 6],
            [7, 8, 9],
        ],
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert set(df.columns) == set(["a", "b", "c"])


def test_csv_3():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        lines=[
            [1, "x", 3],
            [4, "y", 6],
            [7, "z", 9],
        ],
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert set(df.columns) == set(["a", "b", "c"])


def test_csv_4():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        quote_strings=True,
        lines=[
            [1, "x", 3],
            [4, "y", 6],
            [7, "z", 9],
        ],
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert set(df.columns) == set(["a", "b", "c"])


def test_csv_5():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        quote_strings=True,
        lines=[
            [1, "x", 3],
            [4, "y", 6],
            [7, "z", 9],
        ],
    )

    df = s.to_pandas(pandas_read_csv_kwargs={"index_col": "a"})
    assert df.index.name == "a"


@pytest.mark.skipif(True, reason="Test not yet implemented")
def test_csv_icoads():
    r = {
        "class": "e2",
        "date": "1662-10-01/to/1663-12-31",
        "dataset": "icoads",
        "expver": "1608",
        "groupid": "17",
        "reportype": "16008",
        "format": "ascii",
        "stream": "oper",
        "time": "all",
        "type": "ofb",
    }

    source = from_source("mars", **r)
    print(source)


def test_csv_text_file():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        quote_strings=True,
        lines=[
            [1, "x", 3],
            [4, "y", 6],
            [7, "z", 9],
        ],
        extension=".txt",
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert set(df.columns) == set(["a", "b", "c"])


def test_csv_with_comment():
    s = from_source(
        "dummy-source",
        "csv",
        headers=["a", "b", "c"],
        quote_strings=True,
        lines=[
            [1, "x", 3],
            [4, "y", 6],
            [7, "z", 9],
        ],
        comment_line="This is a comment",
    )

    df = s.to_pandas()
    assert len(df) == 3
    assert set(df.columns) == set(["a", "b", "c"])

    ds = s.to_xarray()
    assert len(ds) == 3
    assert set(ds.variables) == set(["index", "a", "b", "c"])


def test_csv_mimetypes():
    assert mimetypes.guess_type("x.csv") == ("text/csv", None)
    assert mimetypes.guess_type("x.csv.gz") == ("text/csv", "gzip")
    assert mimetypes.guess_type("x.csv.bz2") == ("text/csv", "bzip2")


# TODO test compression

if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def test_empty_source_len():
    ds = from_source("empty")
    assert len(ds) == 0


def test_empty_source_concat1():
    ds = from_source("empty")
    assert len(ds) == 0

    ds1 = from_source("file", earthkit_examples_file("test.grib"))

    ds2 = ds + ds1
    assert len(ds2) == 2
    meta = ds2.metadata("shortName")
    assert meta == ["2t", "msl"]


def test_empty_source_concat2():
    ds = from_source("empty")
    assert len(ds) == 0

    ds1 = from_source("file", earthkit_examples_file("test.grib"))

    ds = ds + ds1
    assert len(ds) == 2
    meta = ds.metadata("shortName")
    assert meta == ["2t", "msl"]


def test_empty_source_concat3():
    ds = from_source("empty")
    assert len(ds) == 0

    ds1 = from_source("file", earthkit_examples_file("test.grib"))

    ds += ds1
    assert len(ds) == 2
    meta = ds.metadata("shortName")
    assert meta == ["2t", "msl"]


def test_empty_source_iterate():
    ds = from_source("empty")
    for _ in ds:
        assert False, "Empty source should not iterate"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pandas as pd
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file

assert_frame_equal = pd.testing.assert_frame_equal


@pytest.mark.parametrize("_kwargs", [{}, {"filters": {}}])
def test_bufr_to_pandas_no_filters(_kwargs):
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))

    res = ds.to_pandas(columns=["latitude", "longitude", "WMO_station_id"], **_kwargs)

    row_ref_0 = {"latitude": 67.37, "longitude": 26.63, "WMO_station_id": 2836}
    row_ref_9 = {"latitude": 60.82, "longitude": 23.50, "WMO_station_id": 2963}

    ref_0 = pd.DataFrame(row_ref_0, index=[0])
    ref_9 = pd.DataFrame(row_ref_9, index=[9])

    assert len(res) == 10
    assert list(res.columns) == ["latitude", "longitude", "WMO_station_id"]
    assert_frame_equal(res[0:1], ref_0[res.columns])
    assert_frame_equal(res[9:10], ref_9[res.columns])


def test_bufr_to_pandas_filters():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))

    res = ds.to_pandas(
        columns=["latitude", "longitude", "WMO_station_id"],
        filters={"WMO_station_id": [2836, 2963]},
    )

    ref = pd.DataFrame.from_dict(
        {
            "latitude": [67.37, 60.82],
            "longitude": [26.63, 23.50],
            "WMO_station_id": [2836, 2963],
        }
    )
    assert_frame_equal(res, ref[res.columns])


def test_bufr_to_pandas_multi():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr")) + from_source(
        "file", earthkit_examples_file("synop_10.bufr")
    )

    res = ds.to_pandas(
        columns=["latitude", "longitude", "WMO_station_id"],
    )

    assert len(res) == 20


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()
